import numpy as np
import wx
import wx.adv
import wx.lib.filebrowsebutton as filebrowse

from genx.core.custom_logging import iprint

from ..data_loader_framework import Template
from ..utils import ShowErrorDialog, ShowWarningDialog
from .help_modules import spec

_maxWidth = 450


class Plugin(Template):
    def __init__(self, parent):
        Template.__init__(self, parent)
        self.specfile = None
        self.specfile_name = None
        self.scan = None
        # self.dataset = data.DataSet()
        self.datalist = self.data

    def LoadDataFile(self, selected_items, data_id=0):
        """Called when GenX wants to load data"""

        if len(selected_items) == 0:
            ShowWarningDialog(self.parent, "Please select a data set before trying to load a spec file.")
            return False
        old_data = self.datalist[selected_items[0]].copy()
        self.dataset = self.datalist[selected_items[0]]
        wizard = wx.adv.Wizard(self.parent, -1, "Load Spec File", style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
        # page1 = TitledPage(wizard, "Page1")
        page1 = LoadSpecScanPage(wizard, self, default_filename=self.specfile_name)
        page2 = SelectCountersPage(wizard, self)
        page3 = CustomManipulationPage(wizard, self)
        page4 = SetNamePage(wizard, self)

        wx.adv.WizardPageSimple_Chain(page1, page2)
        wx.adv.WizardPageSimple_Chain(page2, page3)
        wx.adv.WizardPageSimple_Chain(page3, page4)

        wizard.FitToPage(page1)
        # print dir(wizard)
        # print 'data: ', self.datalist[0].x
        if wizard.RunWizard(page1):
            return True
        else:
            self.datalist[selected_items[0]].safe_copy(old_data)
            return False

    def load_specfile(self, filename):
        self.specfile = spec.SpecDataFile(filename)
        names = list(self.specfile.findex.keys())
        names.sort()
        try:
            sc = self.specfile.scan_commands
            names = ["%s " % (name,) + sc[name] for name in names]
        except Exception as e:
            iprint("Could not create full names, error: ", e.__str__())
            ShowErrorDialog(self.parent, "Could not create full names, error:\n %s" % e.__str__())
            names = ["%s " % name for name in names]
        self.specfile_name = filename
        self.scan_names = names
        return self.scan_names

    def load_scans(self, scanlist):
        """Function to load the spec scans as a list of scan numbers"""
        # self.dataset = data.DataSet()
        try:
            self.scan = self.specfile[scanlist]
            self.dataset.set_extra_data("values", self.scan.values)
        except Exception as e:
            iprint("Could not load the desired scans")
            iprint("Error: ", e.__str__())
            iprint("scanlist: ", scanlist)
            ShowErrorDialog(self.parent, "Could not load the desired scans, error:\n%s " % e.__str__())
        # for key in self.scan.values:
        #    try:
        #        self.dataset.set_extra_data(key, self.scan.values[key])
        #    except Exception, e:
        #        print "Could not load ", key
        #        print error

    def get_data_choices(self):
        """Function to return the data choices"""
        try:
            choices = self.scan.cols
        except Exception as e:
            iprint("Could not load the scan cols error: ", e.__str__())
            ShowErrorDialog(self.parent, "Could not load the scan cols, error:\n%s " % e.__str__())
            choices = []

        return choices

    def update_data_cols(self, cols, autom=False):
        """Update the choices for the different values"""
        iprint(cols)
        self.x_val = cols[0]
        self.det_val = cols[1]
        self.mon_val = cols[2]
        self.error_val = cols[3]
        xval = self.scan.values[self.x_val]
        yvals = [
            self.scan.values[self.det_val],
        ]
        if self.mon_val != "None":
            yvals.append(self.scan.values[self.mon_val])
        if self.error_val != "None" and self.error_val != "":
            yvals.append(self.scan.values[self.error_val])
        if autom:
            xval, yvals = automerge(xval, yvals)
        # self.dataset.set_extra_data('det', self.scan.values[self.det_val])
        self.dataset.set_extra_data("det", yvals[0])
        if self.mon_val != "None":
            # self.dataset.set_extra_data('mon', self.scan.values[self.mon_val])
            self.dataset.set_extra_data("mon", yvals[1])

            if self.error_val != "None" and self.error_val != "":
                self.dataset.error_raw = yvals[2]
        elif self.error_val != "None" and self.error_val != "":
            self.dataset.error_raw = yvals[1]

        # self.dataset.x_raw = self.scan.values[self.x_val]
        self.dataset.x_raw = xval
        # if self.error_val != 'None' and self.error_val != '':
        #    self.dataset.error_raw = self.scan.values[self.error_val]

    def get_dataset_names(self):
        return [d.name for d in self.datalist]

    def set_commands(self, commands):
        """Sets the data commands if valid to commands"""
        result = self.dataset.try_commands(commands)
        if result != "":
            ShowErrorDialog(self.parent, result)
            return False
        self.dataset.set_commands(commands)
        self.dataset.run_command()
        return True


class TitledPage(wx.adv.WizardPageSimple):
    def __init__(self, parent, title, plugin):
        wx.adv.WizardPageSimple.__init__(self, parent)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)
        titleText = wx.StaticText(self, -1, title)
        titleText.SetFont(wx.Font(18, wx.SWISS, wx.NORMAL, wx.BOLD))
        self.sizer.Add(titleText, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        self.sizer.Add(wx.StaticLine(self, -1), 0, wx.EXPAND | wx.ALL, 5)


class LoadSpecScanPage(wx.adv.WizardPageSimple):

    def __init__(self, parent, plugin, default_filename=None):
        self.plugin = plugin

        title = "Load scan"
        wx.adv.WizardPageSimple.__init__(self, parent)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)
        titleText = wx.StaticText(self, -1, title)
        titleText.SetFont(wx.Font(18, wx.SWISS, wx.NORMAL, wx.BOLD))
        self.sizer.Add(titleText, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        self.sizer.Add(wx.StaticLine(self, -1), 0, wx.EXPAND | wx.ALL, 5)

        self.filebrowser = filebrowse.FileBrowseButton(
            self, -1, startDirectory=".", size=(450, -1), changeCallback=self.filebrowser_callback, fileMode=wx.FD_OPEN
        )
        self.filebrowser.textControl.Disable()
        self.sizer.Add(self.filebrowser, 0, wx.EXPAND | wx.ALL, 10)
        chooserText = wx.StaticText(self, -1, "Select the scans to load")
        self.sizer.Add(chooserText, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        self.scanchooser = wx.CheckListBox(self, -1)
        self.sizer.Add(self.scanchooser, 1, wx.EXPAND | wx.ALL, 2)

        self.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGING, self.OnWizPageChanging)
        self.scanchooser.Bind(wx.EVT_CHECKLISTBOX, self.OnScanSelected)

        self.SetControlEnable(wx.ID_FORWARD, False)

        # Load the defualt spec file -  if exists
        if default_filename:
            self.filebrowser.SetValue(default_filename)
            self.load_file(default_filename)

    def SetControlEnable(self, id, state):
        """Sets the control enable of window for parent to state. Use wx.ID_FORWARD or
        wx.ID_BACKWARD for the beck forward buttons.
        """
        win = self.Parent.FindWindowById(id)
        if win:
            win.Enable(state)
        else:
            raise Exception("Can not find window %s" % (id,))

    def OnScanSelected(self, evt):
        """A scan has been seelcted"""
        # check so at least one scan is selected
        if len(self.scanchooser.GetChecked()) > 0:
            self.SetControlEnable(wx.ID_FORWARD, True)
        else:
            self.SetControlEnable(wx.ID_FORWARD, False)

    def OnWizPageChanging(self, evt):
        # Check if we are moving forward
        if evt.GetDirection():
            scan_numbers = self.extract_scan_numbers()
            if self.plugin.specfile:
                try:
                    self.plugin.load_scans(scan_numbers)
                except Exception as e:
                    ShowErrorDialog(
                        self.Parent,
                        "Could not load the selected scan"
                        " - The scan might be corrupt.\n"
                        "Please choose another.\n"
                        "Python error: " + e.__str__(),
                    )
                    # print e.__str__()
                    evt.Veto()

    def extract_scan_numbers(self):
        """Extract the scan numbers from the Checklistbox"""
        scan_numbers = self.scanchooser.CheckedStrings
        # Extract the numbers from the strings
        scan_numbers = [int(ss.split(" ")[0]) for ss in scan_numbers]
        return scan_numbers

    def load_file(self, filename):
        self.scanchooser.Set(self.plugin.load_specfile(filename))
        # self.SetControlEnable(wx.ID_FORWARD, True)
        # print filename

    def filebrowser_callback(self, evt):
        filename = evt.GetString()
        self.load_file(filename)


class SelectCountersPage(wx.adv.WizardPageSimple):
    def __init__(self, parent, plugin):
        wx.adv.WizardPageSimple.__init__(self, parent)
        self.plugin = plugin
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)
        title = "Select Counters/Motors"
        titleText = wx.StaticText(self, -1, title)
        titleText.SetFont(wx.Font(18, wx.SWISS, wx.NORMAL, wx.BOLD))
        self.sizer.Add(titleText, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        self.sizer.Add(wx.StaticLine(self, -1), 0, wx.EXPAND | wx.ALL, 5)

        self.choiceSizer = wx.FlexGridSizer(rows=4, cols=2, vgap=10, hgap=10)
        self.choiceSizer.Add(wx.StaticText(self, -1, "X value"), 0, wx.EXPAND | wx.ALL, 5)
        self.xChoice = wx.Choice(self, -1, (-1, -1), choices=[])
        self.choiceSizer.Add(self.xChoice, 0, wx.EXPAND | wx.ALL, 5)
        self.choiceSizer.Add(wx.StaticText(self, -1, "Det. value"), 0, wx.EXPAND | wx.ALL, 5)
        self.detChoice = wx.Choice(self, -1, (-1, -1), choices=[])
        self.choiceSizer.Add(self.detChoice, 0, wx.EXPAND | wx.ALL, 5)
        self.choiceSizer.Add(wx.StaticText(self, -1, "Mon. value"), 0, wx.EXPAND | wx.ALL, 5)
        self.monChoice = wx.Choice(self, -1, (-1, -1), choices=[])
        self.choiceSizer.Add(self.monChoice, 0, wx.EXPAND | wx.ALL, 5)
        self.choiceSizer.Add(wx.StaticText(self, -1, "Error value"), 0, wx.EXPAND | wx.ALL, 5)
        self.errorChoice = wx.Choice(self, -1, (-1, -1), choices=[])
        self.choiceSizer.Add(self.errorChoice, 0, wx.EXPAND | wx.ALL, 5)
        self.sizer.Add(self.choiceSizer, 0, wx.ALIGN_CENTRE | wx.ALL)
        self.automergeCheckBox = wx.CheckBox(self, -1, "Auto merge data")
        self.automergeSizer = wx.FlexGridSizer(rows=1, cols=1, vgap=10, hgap=10)
        self.automergeSizer.Add(self.automergeCheckBox, 0, wx.EXPAND | wx.ALL, 5)
        self.sizer.Add(self.automergeSizer, 0, wx.ALIGN_CENTRE | wx.ALL)

        self.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGED, self.OnWizPageChanged)
        self.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGING, self.OnWizPageChanging)

    def OnWizPageChanged(self, evt):
        # Check if we are moving forward
        # print 'In wiz page changing ', evt.GetDirection()
        if evt.GetDirection():
            choices = self.plugin.get_data_choices()
            self.update_counters(choices)

    def OnWizPageChanging(self, evt):
        if evt.GetDirection():
            self.plugin.update_data_cols(self.get_selected(), self.automergeCheckBox.IsChecked())

    def update_counters(self, mot_count):
        self.xChoice.SetItems(mot_count)
        if len(mot_count) > 0:
            self.xChoice.SetSelection(0)
        self.detChoice.SetItems(mot_count)
        try:
            self.detChoice.SetSelection(mot_count.index("Detector"))
        except:
            self.detChoice.SetSelection(0)
        self.monChoice.SetItems(["None"] + mot_count)
        try:
            self.monChoice.SetSelection(mot_count.index("Monitor") + 1)
        except:
            self.monChoice.SetSelection(0)
        self.errorChoice.SetItems(["None"] + mot_count)
        self.errorChoice.SetSelection(0)

    def get_selected(self):
        data = (
            self.xChoice.GetStringSelection(),
            self.detChoice.GetStringSelection(),
            self.monChoice.GetStringSelection(),
            self.errorChoice.GetStringSelection(),
        )
        return data


class CustomManipulationPage(wx.adv.WizardPageSimple):
    def __init__(self, parent, plugin, data_sets=None):
        wx.adv.WizardPageSimple.__init__(self, parent)
        if data_sets is None:
            data_sets = []
        self.plugin = plugin
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)
        title = "Custom Manipulation"
        titleText = wx.StaticText(self, -1, title)
        titleText.SetFont(wx.Font(18, wx.SWISS, wx.NORMAL, wx.BOLD))
        self.sizer.Add(titleText, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        self.sizer.Add(wx.StaticLine(self, -1), 0, wx.EXPAND | wx.ALL, 5)
        help_text = (
            "By changing the expressions in the boxes below you can "
            + "manipulate your data with expressions written in Python.\n"
            + "These transforms can be changed later from the data calculations"
            + "dialog."
        )
        helpText = wx.StaticText(self, -1, help_text)
        helpText.Wrap(_maxWidth)
        self.sizer.Add(helpText, 0, wx.ALIGN_CENTRE | wx.ALL, 10)
        self.manipSizer = wx.FlexGridSizer(rows=4, cols=2, vgap=10, hgap=10)
        self.manipSizer.Add(wx.StaticText(self, -1, "x = "), 0, wx.EXPAND | wx.ALL, 5)
        self.xCtrl = wx.TextCtrl(self, -1, value="x")
        self.manipSizer.Add(self.xCtrl, 0, wx.EXPAND | wx.ALL, 5)

        self.manipSizer.Add(wx.StaticText(self, -1, "y = "), 0, wx.EXPAND | wx.ALL, 5)
        self.yCtrl = wx.TextCtrl(self, -1, value="det")
        self.manipSizer.Add(self.yCtrl, 0, wx.EXPAND | wx.ALL, 5)

        self.manipSizer.Add(wx.StaticText(self, -1, "Error = "), 0, wx.EXPAND | wx.ALL, 5)
        self.errorCtrl = wx.TextCtrl(self, -1, value="0*x")
        self.manipSizer.Add(self.errorCtrl, 0, wx.EXPAND | wx.ALL, 5)
        self.sizer.Add(self.manipSizer, 0, wx.ALIGN_CENTRE | wx.ALL)

        help_text = "Import the settings from data set: "
        helpText = wx.StaticText(self, -1, help_text)
        helpText.Wrap(_maxWidth)
        self.sizer.Add(helpText, 0, wx.ALIGN_CENTRE | wx.ALL, 10)
        self.importChoice = wx.Choice(self, -1, (-1, 50), choices=data_sets)
        self.sizer.Add(self.importChoice, 0, wx.ALIGN_CENTRE | wx.ALL, 10)
        self.Bind(wx.EVT_CHOICE, self.EvtImport, self.importChoice)

        self.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGED, self.OnWizPageChanged)
        self.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGING, self.OnWizPageChanging)

    def OnWizPageChanged(self, evt):
        """Event handler for entering this page"""
        manip_string = ["x", "det"]
        if self.plugin.mon_val != "None":
            manip_string.append("mon")
        else:
            manip_string.append("None")
        if self.plugin.error_val != "None":
            manip_string.append("e")
        else:
            manip_string.append("None")

        self.set_manip(manip_string)
        self.importChoice.SetItems(self.plugin.get_dataset_names())

    def OnWizPageChanging(self, evt):
        """Event Handler for leaving this page"""
        commands = {
            "x": self.xCtrl.GetValue(),
            "y": self.yCtrl.GetValue(),
            "e": self.errorCtrl.GetValue(),
        }
        if not self.plugin.set_commands(commands):
            evt.Veto()

    def EvtImport(self, evt):
        """On the event to import manipulations from another data set"""
        item = self.importChoice.GetSelection()
        com = self.plugin.datalist[item].get_commands()
        # self.set_manip((com['x'], com['y'], com['e']))
        self.xCtrl.ChangeValue(com["x"])
        self.yCtrl.ChangeValue(com["y"])
        self.errorCtrl.ChangeValue(com["e"])

    def set_manip(self, choices):
        """Sets the manipulations to defualt values as given by
        the choices in the form (x, det, mon, error)
        """
        iprint(choices)
        self.xCtrl.ChangeValue(choices[0])
        if choices[2] == "None":
            self.yCtrl.ChangeValue(choices[1])
            if choices[3] == "None":
                self.errorCtrl.ChangeValue("sqrt(%s)" % (choices[1],))
            else:
                self.errorCtrl.ChangeValue(choices[3])
        else:
            self.yCtrl.ChangeValue("%s/%s" % (choices[1], choices[2]))
            if choices[3] == "None":
                self.errorCtrl.ChangeValue(
                    "sqrt(1.0/%s + 1.0/%s)*%s/%s" % (choices[1], choices[2], choices[1], choices[2])
                )
            else:
                self.errorCtrl.ChangeValue(choices[3])


class SetNamePage(wx.adv.WizardPageSimple):

    def __init__(self, parent, plugin, data_sets=None):
        wx.adv.WizardPageSimple.__init__(self, parent)
        if data_sets is None:
            data_sets = []
        self.plugin = plugin
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)
        title = "Name"
        titleText = wx.StaticText(self, -1, title)
        titleText.SetFont(wx.Font(18, wx.SWISS, wx.NORMAL, wx.BOLD))
        self.sizer.Add(titleText, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        self.sizer.Add(wx.StaticLine(self, -1), 0, wx.EXPAND | wx.ALL, 5)
        help_text = (
            "Set the name of the data set as it will appear in GenX."
            "The name can be changed later by a slow double click on the name "
            "in the data panel."
        )
        helpText = wx.StaticText(self, -1, help_text)
        helpText.Wrap(_maxWidth)
        self.sizer.Add(helpText, 0, wx.ALIGN_CENTRE | wx.ALL, 10)

        self.manipSizer = wx.FlexGridSizer(rows=1, cols=2, vgap=10, hgap=10)
        self.manipSizer.Add(wx.StaticText(self, -1, "Name: "), 0, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL | wx.ALL, 10)
        self.xCtrl = wx.TextCtrl(self, -1, value="Name")
        self.manipSizer.Add(self.xCtrl, 0, wx.EXPAND | wx.ALL, 10)
        self.sizer.Add(self.manipSizer, 0, wx.EXPAND | wx.ALIGN_CENTRE | wx.ALL)

        self.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGING, self.OnWizPageChanging)

    def OnWizPageChanging(self, evt):
        """Event Handler for leaving this page"""

        self.plugin.dataset.name = self.xCtrl.GetValue()


def automerge(xval, yvals, rel_cond=100.0):
    """Merges the values according to rel_cond where the
    the merging criterion is the distance between the two first datapoints divided by
    rel_cond.

    xval - an arraray of x values
    y-vals a tuple of yvals
    rel_cond as above
    """
    min_size = np.abs(xval[1] - xval[0]) / rel_cond
    xarg = xval.argsort()
    xnew = np.array([xval[xarg[0]]]) * 1.0
    ynew = []
    for yval in yvals:
        ynew.append(np.array([yval[xarg[0]]]) * 1.0)
    inew = 0
    for i in range(len(xarg))[1:]:
        if np.abs(xnew[-1] - xval[xarg[i]]) <= min_size:
            for j in range(len(ynew)):
                ynew[j][-1] += yvals[j][xarg[i]]
        else:
            xnew = np.append(xnew, xval[xarg[i]])
            for j in range(len(ynew)):
                ynew[j] = np.append(ynew[j], yvals[j][xarg[i]])
            inew += 1

    return xnew, ynew


if __name__ == "__main__":
    app = wx.SimpleApp()
    plugin = Plugin(-1)
    wizard = wx.adv.Wizard(None, -1, "Simple Wizard", style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
    # page1 = TitledPage(wizard, "Page1")
    page2 = LoadSpecScanPage(wizard, plugin)
    page3 = SelectCountersPage(wizard, plugin)
    page4 = CustomManipulationPage(wizard, plugin)
    page5 = SetNamePage(wizard, plugin)

    # page1.sizer.Add(wx.StaticText(page1, -1, "Testing the wizard"))
    # page4.sizer.Add(wx.StaticText(page4, -1, "This is the last page"))

    # wx.adv.WizardPageSimple_Chain(page1, page2)
    wx.adv.WizardPageSimple_Chain(page2, page3)
    wx.adv.WizardPageSimple_Chain(page3, page4)
    wx.adv.WizardPageSimple_Chain(page4, page5)

    wizard.FitToPage(page2)
    # print dir(wizard)
    if wizard.RunWizard(page2):
        iprint("Sucess")



from genx.version import __version__

mv = int(__version__.split(".")[1])

with open("README_latest.txt", "w") as fh:
    pc = f"Changes 3.{mv-1}"
    for l in open("README.txt", "r").readlines():
        if pc in l:
            break
        fh.write(l)



"""
Library for combined x-ray and neutrons simulations.
====================================================
The neutron simulations is capable of handling non-magnetic, magnetic
non-spin flip as well as neutron spin-flip reflectivity.

Classes
-------

"""

from dataclasses import dataclass, field, fields
from typing import List

import numpy as np

# import all special footprint functions
from .lib import footprint as footprint_module
from .lib import neutron_refl as MatrixNeutron
from .lib import paratt as Paratt
from .lib import refl_base as refl
from .lib import resolution as resolution_module
from .lib.base import AltStrEnum
from .lib.footprint import *
from .lib.instrument import *
from .lib.physical_constants import AA_to_eV, muB_to_SL, r_e
from .lib.resolution import *
from .lib.testing import ModelTestCase

# Preamble to define the parameters needed for the models outlined below:

ModelID = "SpecNX"

q_limit = 1e-10
""" Minimum allowed q-value """

__xlabel__ = "q [Å$^{-1}$]"
__ylabel__ = "Instnsity [a.u.]"


class Probe(AltStrEnum):
    xray = "x-ray"
    neutron = "neutron"
    npol = "neutron pol"
    npolsf = "neutron pol spin flip"
    ntof = "neutron tof"
    ntofpol = "neutron pol tof"


class Coords(AltStrEnum):
    q = "q"
    tth = "2θ"
    alternate_tth = "tth"


class ResType(AltStrEnum):
    none = "no conv"
    fast_conv = "fast conv"
    fast_conv_var = "fast conv + varying res."
    fast_conv_rel = "fast conv + varying res. (dx/x)"
    full_conv_abs = "full conv and varying res."
    full_conv_rel = "full conv and varying res. (dx/x)"


class FootType(AltStrEnum):
    none = "no corr"
    gauss = "gauss beam"
    square = "square beam"


class Polarization(AltStrEnum):
    up_up = "uu"
    down_down = "dd"
    up_down = "ud"
    down_up = "du"
    asymmetry = "ass"

    alternate_up_up = "++"
    alternate_down_down = "--"
    alternate_up_down = "+-"
    alternate_down_up = "-+"


@dataclass
class Layer(refl.ReflBase):
    """
    Representing a layer in the sample structur.

    ``d``
        The thickness of the layer in AA (Angstroms = 1e-10m)
    ``dens``
        The density of formula units in units per Angstroms. Note the units!
    ``sigma``
        The root mean square roughness of the top interface of the layer in Angstroms.
    ``f``
       The x-ray scattering length per formula unit in electrons. To be
       strict it is the number of Thompson scattering lengths for each
       formula unit.
    ``b``
       The neutron scattering length per formula unit in fm (femtometer = 1e-15m)
    ``xs_ai``
       The sum of the absorption cross section and the incoherent scattering
       cross section in barns for neutrons
    ``magn``
        The magnetic moment per formula unit (same formula unit as b and dens refer to)
    ``magn_ang``
        The angle of the magnetic moment in degress. 0 degrees correspond to
        a moment collinear with the neutron spin. Only used if probe is neutron spin-flip.
    """

    d: float = 0.0
    dens: float = 1.0
    sigma: float = 0.0

    f: complex = 1e-20j

    b: complex = 0j
    xs_ai: float = 0.0
    magn: float = 0.0
    magn_ang: float = 0.0

    Units = {
        "sigma": "AA",
        "dens": "at./AA",
        "d": "AA",
        "f": "el./at.",
        "b": "fm/at.",
        "xs_ai": "barn/at.",
        "magn": "mu_B/at.",
        "magn_ang": "deg.",
    }

    Groups = [("General", ["d", "dens", "sigma"]), ("Neutron", ["b", "xs_ai", "magn", "magn_ang"]), ("X-Ray", ["f"])]

    DEFAULT_FIT_PARAMS = ["d", "dens", "sigma"]


@dataclass
class LayerParameters:
    d: List[float]
    dens: List[float]
    sigma: List[float]

    f: List[complex]

    b: List[complex]
    xs_ai: List[float]
    magn: List[float]
    magn_ang: List[float]


@dataclass
class Stack(refl.StackBase):
    """
    A collection of Layer objects that can be repeated.

    ``Layers``
       A ``list`` consiting of ``Layers`` in the stack the first item is
       the layer closest to the bottom
    ``Repetitions``
       The number of repsetions of the stack
    """

    Layers: List[Layer] = field(default_factory=list)
    Repetitions: int = 1


@dataclass
class Sample(refl.SampleBase):
    """
    Describe global sample by listing ambient, substrate and layer parameters.

    ``Stacks``
       A ``list`` consiting of ``Stacks`` in the stacks the first item is
       the layer closest to the bottom
    ``Ambient``
       A ``Layer`` describing the Ambient (enviroment above the sample).
       Only the scattering lengths and density of the layer is used.
    ``Substrate``
       A ``Layer`` describing the substrate (enviroment below the sample).
       Only the scattering lengths, density and roughness of the layer is
       used.
    ``crop_sld``
       For samples with many layers this limits the number of layers that
       are shown on SLD graphs. Negative values just remove these layers
       and show a spike in SLD, instead, while postivie values exchange
       them by one layer of same thickness.
    """

    Stacks: List[Stack] = field(default_factory=list)
    Ambient: Layer = field(default_factory=Layer)
    Substrate: Layer = field(default_factory=Layer)
    crop_sld: int = 0
    _layer_parameter_class = LayerParameters


@dataclass
class Instrument(refl.ReflBase):
    """
    Specify parameters of the probe and reflectometry instrument.

    ``probe``
        Describes the radiation and measurments used, it is one of:
        'x-ray', 'neutron', 'neutron pol', 'neutron pol spin flip',
        'neutron tof', 'neutron pol tof'.
        The calculations for x-rays uses ``f`` for the scattering length for
        neutrons ``b`` for 'neutron pol', 'neutron pol spin flip' and 'neutron
        pol tof' alternatives the ``magn`` is used in the calculations. Note
        that the angle of magnetization ``magn_ang`` is only used in the spin
        flip model.
    ``wavelength``
        The wavelength of the radiation given in AA (Angstroms)
    ``I0``
        The incident intensity (a scaling factor)
    ``Ibkg``
        The background intensity. Added as a constant value to the calculated
        reflectivity
    ``pol``
        The measured polarization of the instrument. Valid options are:
        'uu','dd', 'ud', 'du' or 'ass' the respective number 0-3 also works.
    ``coords``
        The coordinates of the data given to the SimSpecular function. The
        available alternatives are: 'q' or '2θ'. Alternatively the numbers 0 (q)
        or 1 (tth) can be used.
    ``tthoff``
        Linear offset to the scattering angle calibration
    ``incangle``
        The incident angle of the neutrons, only valid in tof mode
    ``restype``
        Describes the rype of the resolution calculated. One of the
        alterantives: 'no conv', 'fast conv', 'full conv and varying res.',
        'fast conv + varying res.', 'full conv and varying res. (dx/x)', 'fast
        conv + varying res. (dx/x)'. The respective numbers 0-3 also works. Note
        that fast convolution only alllows a single value into res wheras the
        other can also take an array with the same length as the x-data (varying
        resolution)
    ``res``
        The resolution of the instrument given in the coordinates of ``coords``.
        This assumes a gaussian resolution function and ``res`` is the standard
        deviation of that gaussian. If ``restype`` has (dx/x) in its name the
        gaussian standard deviation is given by res*x where x is either in tth
        or q.
    ``respoints``
        The number of points to include in the resolution calculation. This is
        only used for 'full conv and vaying res.', 'fast conv + varying res',
        'full conv and varying res. (dx/x)' and 'fast conv + varying res.
        (dx/x)'.
    ``resintrange``
        Number of standard deviatons to integrate the resolution function times
        the reflectivity over
    ``footype``
        Which type of footprint correction is to be applied to the simulation.
        One of: 'no corr', 'gauss beam' or 'square beam'. Alternatively, the
        number 0-2 are also valid. The different choices are self expnalatory.
    ``beamw``
        The width of the beam given in mm. For 'gauss beam' it should be the
        standard deviation. For 'square beam' it is the full width of the beam.
    ``samplelen``
        The length of the sample given in mm
    """

    probe: Probe = "x-ray"
    wavelength: float = 1.54
    I0: float = 1.0
    Ibkg: float = 0.0
    pol: Polarization = "uu"
    coords: Coords = "2θ"
    tthoff: float = 0.0
    incangle: float = 0.5
    restype: ResType = "no conv"
    res: float = 0.001
    respoints: int = 5
    resintrange: float = 2.0
    footype: FootType = "no corr"
    beamw: float = 0.01
    samplelen: float = 10.0

    Units = {
        "probe": "",
        "wavelength": "Å",
        "coords": "",
        "I0": "arb.",
        "res": "[coord]",
        "restype": "",
        "respoints": "pts.",
        "resintrange": "[coord]",
        "beamw": "mm",
        "footype": "",
        "samplelen": "mm",
        "incangle": "°",
        "pol": "",
        "Ibkg": "arb.",
        "tthoff": "°",
        "2θ": "°",
        "tth": "°",
        "q": "Å$^-1$",
    }

    Groups = [
        ("Radiation", ["probe", "wavelength", "I0", "Ibkg", "pol"]),
        ("X-Resolution", ["restype", "res", "respoints", "resintrange"]),
        ("X-Coordinates", ["coords", "tthoff", "incangle"]),
        ("Footprint", ["footype", "beamw", "samplelen"]),
    ]

    DEFAULT_FIT_PARAMS = ["I0", "Ibkg", "Beamw", "res"]

# A buffer to save previous calculations for spin-flip calculations
class Buffer:
    Ruu = 0
    Rdd = 0
    Rdu = 0
    Rud = 0
    parameters = None
    TwoThetaQz = None


def footprintcorr(Q, instrument: Instrument):
    foocor = 1.0
    footype = instrument.footype
    beamw = instrument.beamw
    samlen = instrument.samplelen

    if instrument.probe in [Probe.ntof, Probe.ntofpol]:
        if type(instrument.incangle) is float:
            theta = instrument.incangle * ones_like(Q)
        else:
            theta = instrument.incangle
        # if ai is an array, make sure it gets repeated for every resolution point
        if type(theta) is ndarray and instrument.restype in [ResType.full_conv_rel, ResType.full_conv_abs]:
            theta = (theta * ones(instrument.respoints)[:, newaxis]).flatten()

    else:
        # theta =  180./pi*arcsin(instrument.wavelength / 4.0 / pi * Q)
        theta = QtoTheta(instrument.wavelength, Q)
    if footype == FootType.gauss:
        foocor = GaussIntensity(theta, samlen / 2.0, samlen / 2.0, beamw)
    elif footype == FootType.square:
        foocor = SquareIntensity(theta, samlen, beamw)
    elif footype == FootType.none:
        pass
    elif isinstance(footype, footprint_module.Footprint):
        foocor = footype(theta, samlen)
    else:
        raise ValueError("The choice of footprint correction, footype," "is WRONG")

    return foocor


def resolutioncorr(R, TwoThetaQz, foocor, instrument: Instrument, weight):
    """Do the convolution of the reflectivity to account for resolution effects."""
    R = R[:] * foocor
    restype = instrument.restype
    if restype == ResType.fast_conv:
        R = ConvoluteFast(TwoThetaQz, R, instrument.res, range=instrument.resintrange)
    elif restype in [ResType.full_conv_rel, ResType.full_conv_abs] or isinstance(restype, resolution_module.Resolution):
        R = ConvoluteResolutionVector(TwoThetaQz, R, weight)
    elif restype == ResType.fast_conv_var:
        R = ConvoluteFastVar(TwoThetaQz, R, instrument.res, range=instrument.resintrange)
    elif restype == ResType.fast_conv_rel:
        R = ConvoluteFastVar(TwoThetaQz, R, instrument.res * TwoThetaQz, range=instrument.resintrange)
    elif restype == ResType.none:
        pass
    else:
        raise ValueError("The choice of resolution type, restype," "is WRONG")
    return R


def resolution_init(TwoThetaQz, instrument: Instrument):
    """Inits the dependet variable with regards to coordinates and resolution."""
    global __xlabel__
    __xlabel__ = "q [Å$^{-1}$]"

    restype = instrument.restype
    weight = 0
    if isinstance(restype, resolution_module.Resolution):
        (TwoThetaQz, weight) = restype(TwoThetaQz[:], instrument.respoints, instrument.resintrange)
    elif restype == ResType.full_conv_abs:
        res_function = GaussianResolution(sigma=instrument.res)
        (TwoThetaQz, weight) = res_function(TwoThetaQz[:], instrument.respoints, instrument.resintrange)
    elif restype == ResType.full_conv_rel:
        res_function = GaussianResolution(sigma=instrument.res * TwoThetaQz)
        (TwoThetaQz, weight) = res_function(TwoThetaQz[:], instrument.respoints, instrument.resintrange)
    # TTH values given as x
    if instrument.coords == Coords.tth:
        Q = TwoThetatoQ(instrument.wavelength, TwoThetaQz + instrument.tthoff)
        __xlabel__ = "2θ [°]"
    # Q vector given....
    elif instrument.coords == Coords.q:
        # if there is no tth offset, nothing to be done for Q
        if instrument.tthoff == 0:
            Q = TwoThetaQz
        # for tof the q-values are not angles but wavelenght, so tth-offset is a scaling factor
        elif instrument.probe in [Probe.ntof, Probe.ntofpol]:
            ai = instrument.incangle
            # if ai is an array, make sure it gets repeated for every resolution point
            if type(ai) is ndarray and restype in [ResType.full_conv_abs, ResType.full_conv_rel]:
                ai = (ai * ones(instrument.respoints)[:, newaxis]).flatten()
            Q = TwoThetaQz * (sin((ai + instrument.tthoff / 2.0) * pi / 180.0) / sin(ai * pi / 180.0))
        else:
            Q = TwoThetatoQ(
                instrument.wavelength, QtoTheta(instrument.wavelength, TwoThetaQz) * 2.0 + instrument.tthoff
            )
    else:
        raise ValueError("The value for coordinates, coords, is WRONG! should be q(0) or tth(1).")
    return Q, TwoThetaQz, weight


def neutron_sld(abs_xs, dens, fb, wl):
    return dens * (wl**2 / 2 / pi * fb - 1.0j * abs_xs * wl / 4 / pi)


def Specular(TwoThetaQz, sample: Sample, instrument: Instrument):
    """Simulate the specular signal from sample when probed with instrument

    # BEGIN Parameters
    TwoThetaQz data.x
    # END Parameters
    """
    return specular_calcs(TwoThetaQz, sample, instrument, return_int=True)


def SpecularField(TwoThetaQz, sample, instrument):
    """Simulate the specular signal from sample when probed with instrument

    # BEGIN Parameters
    TwoThetaQz data.x
    # END Parameters
    """
    return specular_calcs(TwoThetaQz, sample, instrument, return_int=False)


def specular_calcs(TwoThetaQz, sample: Sample, instrument: Instrument, return_int=True):
    """Simulate the specular signal from sample when probed with instrument

    # BEGIN Parameters
    TwoThetaQz data.x
    # END Parameters
    """

    # preamble to get it working with my class interface
    restype = instrument.restype
    Q, TwoThetaQz, weight = resolution_init(TwoThetaQz, instrument)
    # often an issue with resolution etc. so just replace Q values < q_limit
    Q = maximum(Q, q_limit)

    ptype = instrument.probe
    pol = instrument.pol

    parameters: LayerParameters = sample.resolveLayerParameters()

    dens = array(parameters.dens, dtype=float64)
    d = array(parameters.d, dtype=float64)
    magn = array(parameters.magn, dtype=float64)
    # Transform to radians
    magn_ang = array(parameters.magn_ang, dtype=float64) * pi / 180.0

    sigma = array(parameters.sigma, dtype=float64)

    if ptype == Probe.xray:
        # fb = array(parameters['f'], dtype = complex64)
        e = AA_to_eV / instrument.wavelength
        fb = refl.cast_to_array(parameters.f, e).astype(complex128)
        sld = dens * fb * instrument.wavelength**2 / 2 / pi
    else:
        fb = array(parameters.b, dtype=complex128) * 1e-5
        abs_xs = array(parameters.xs_ai, dtype=complex128) * 1e-4**2
        wl = instrument.wavelength
        # sld = dens*(wl**2/2/pi*sqrt(fb**2 - (abs_xs/2.0/wl)**2) -
        #                       1.0J*abs_xs*wl/4/pi)
        sld = neutron_sld(abs_xs, dens, fb, wl)
    # Ordinary Paratt X-rays
    if ptype == Probe.xray:
        R = Paratt.ReflQ(Q, instrument.wavelength, 1.0 - r_e * sld, d, sigma, return_int=return_int)
        # print 2.82e-5*sld
    # Ordinary Paratt Neutrons
    elif ptype == Probe.neutron:
        R = Paratt.ReflQ(Q, instrument.wavelength, 1.0 - sld, d, sigma, return_int=return_int)
    # Ordinary Paratt but with magnetization
    elif ptype == Probe.npol:
        msld = muB_to_SL * magn * dens * instrument.wavelength**2 / 2 / pi
        # Polarization uu or ++
        if pol == Polarization.up_up:
            R = Paratt.ReflQ(Q, instrument.wavelength, 1.0 - sld - msld, d, sigma, return_int=return_int)
        # Polarization dd or --
        elif pol == Polarization.down_down:
            R = Paratt.ReflQ(Q, instrument.wavelength, 1.0 - sld + msld, d, sigma, return_int=return_int)
        elif pol == Polarization.asymmetry:
            Rp = Paratt.ReflQ(Q, instrument.wavelength, 1.0 - sld - msld, d, sigma, return_int=return_int)
            Rm = Paratt.ReflQ(Q, instrument.wavelength, 1.0 - sld + msld, d, sigma, return_int=return_int)
            R = (Rp - Rm) / (Rp + Rm)

        else:
            raise ValueError("The value of the polarization is WRONG." " It should be uu(0) or dd(1)")
    # Spin flip
    elif ptype == Probe.npolsf:
        # Check if we have calcluated the same sample previous:
        if Buffer.TwoThetaQz is not None:
            Q_ok = Buffer.TwoThetaQz.shape == Q.shape
            if Q_ok:
                Q_ok = not (Buffer.TwoThetaQz != Q).any()
        else:
            Q_ok = False
        if Buffer.parameters != parameters or not Q_ok:
            msld = muB_to_SL * magn * dens * instrument.wavelength**2 / 2 / pi
            # renormalize SLDs if ambient layer is not vacuum
            if msld[-1] != 0.0 or sld[-1] != 0:
                msld -= msld[-1]
                sld -= sld[-1]
            sld_p = sld + msld
            sld_m = sld - msld
            Vp = (2 * pi / instrument.wavelength) ** 2 * (
                sld_p * (2.0 + sld_p)
            )  # (1-np**2) - better numerical accuracy
            Vm = (2 * pi / instrument.wavelength) ** 2 * (sld_m * (2.0 + sld_m))  # (1-nm**2)
            (Ruu, Rdd, Rud, Rdu) = MatrixNeutron.Refl(Q, Vp, Vm, d, magn_ang, sigma, return_int=return_int)
            Buffer.Ruu = Ruu
            Buffer.Rdd = Rdd
            Buffer.Rud = Rud
            Buffer.parameters = parameters
            Buffer.TwoThetaQz = Q.copy()
        else:
            pass
        if pol == Polarization.up_up:
            R = Buffer.Ruu
        elif pol == Polarization.down_down:
            R = Buffer.Rdd
        elif pol in [Polarization.up_down, Polarization.down_up]:
            R = Buffer.Rud
        # Calculating the asymmetry ass
        elif pol == Polarization.asymmetry:
            R = (Buffer.Ruu - Buffer.Rdd) / (Buffer.Ruu + Buffer.Rdd + 2 * Buffer.Rud)
        else:
            raise ValueError("The value of the polarization is WRONG." " It should be uu(0), dd(1) or ud(2)")

    # tof
    elif ptype == Probe.ntof:
        ai = instrument.incangle
        # if ai is an array, make sure it gets repeated for every resolution point
        if type(ai) is ndarray and restype in [ResType.full_conv_rel, ResType.full_conv_abs]:
            ai = (ai * ones(instrument.respoints)[:, newaxis]).flatten()
        else:
            ai = ai * ones(Q.shape)
        wl = 4 * pi * sin(ai * pi / 180) / Q
        sld = neutron_sld(abs_xs[:, newaxis], dens[:, newaxis], fb[:, newaxis], wl)
        R = Paratt.Refl_nvary2(ai, wl, 1.0 - sld, d, sigma, return_int=return_int)
    # tof spin polarized
    elif ptype == Probe.ntofpol:
        wl = 4 * pi * sin(instrument.incangle * pi / 180) / Q
        sld = neutron_sld(abs_xs[:, newaxis], dens[:, newaxis], fb[:, newaxis], wl)
        msld = (
            muB_to_SL
            * magn[:, newaxis]
            * dens[:, newaxis]
            * (4 * pi * sin(instrument.incangle * pi / 180) / Q) ** 2
            / 2
            / pi
        )
        # polarization uu or ++
        if pol == Polarization.up_up:
            R = Paratt.Refl_nvary2(
                instrument.incangle * ones(Q.shape),
                (4 * pi * sin(instrument.incangle * pi / 180) / Q),
                1.0 - sld - msld,
                d,
                sigma,
                return_int=return_int,
            )
        # polarization dd or --
        elif pol == Polarization.down_down:
            R = Paratt.Refl_nvary2(
                instrument.incangle * ones(Q.shape),
                (4 * pi * sin(instrument.incangle * pi / 180) / Q),
                1.0 - sld + msld,
                d,
                sigma,
                return_int=return_int,
            )
        # Calculating the asymmetry
        elif pol == Polarization.asymmetry:
            Rd = Paratt.Refl_nvary2(
                instrument.incangle * ones(Q.shape),
                (4 * pi * sin(instrument.incangle * pi / 180) / Q),
                1.0 - sld + msld,
                d,
                sigma,
                return_int=return_int,
            )
            Ru = Paratt.Refl_nvary2(
                instrument.incangle * ones(Q.shape),
                (4 * pi * sin(instrument.incangle * pi / 180) / Q),
                1.0 - sld - msld,
                d,
                sigma,
                return_int=return_int,
            )
            R = (Ru - Rd) / (Ru + Rd)

        else:
            raise ValueError("The value of the polarization is WRONG." " It should be uu(0) or dd(1) or ass")
    else:
        raise ValueError("The choice of probe is WRONG")
    if return_int:
        # FootprintCorrections
        foocor = footprintcorr(Q, instrument)
        # Resolution corrections
        R = resolutioncorr(R, TwoThetaQz, foocor, instrument, weight)

        return R * instrument.I0 + instrument.Ibkg
    else:
        return R


def EnergySpecular(Energy, TwoThetaQz, sample: Sample, instrument: Instrument):
    """Simulate the specular signal from sample when probed with instrument. Energy should be in eV.

    # BEGIN Parameters
    Energy data.x
    TwoThetaQz 3.0
    # END Parameters
    """
    # preamble to get it working with my class interface
    restype = instrument.restype
    # TODO: Fix so that resolution can be included.
    if restype != 0 and restype != ResType.none:
        raise ValueError("Only no resolution is allowed for energy scans.")
    if instrument.probe != Probe.xray:
        raise NotImplementedError("EnergySpecular only implemented for xray, for neutron ToF use Specular.")

    wl = AA_to_eV / Energy
    global __xlabel__
    __xlabel__ = "E [eV]"

    # TTH values given as x
    if instrument.coords == Coords.tth:
        theta = (TwoThetaQz + instrument.tthoff) / 2.0 * ones_like(Energy)
    # Q vector given....
    elif instrument.coords == Coords.q:
        theta = ((arcsin(TwoThetaQz * wl / 4 / pi) * 180.0 / pi) + instrument.tthoff / 2.0) * ones_like(Energy)

    else:
        raise ValueError("The value for coordinates, coords, is WRONG!" "should be q(0) or tth(1).")

    parameters: LayerParameters = sample.resolveLayerParameters()
    fb = refl.cast_to_array(parameters.f, Energy).astype(complex128)

    dens = array(parameters.dens, dtype=float64)
    d = array(parameters.d, dtype=float64)
    sigma = array(parameters.sigma, dtype=float64)

    sld = (dens * fb)[:, newaxis] * wl**2 / 2 / pi

    R = Paratt.Refl_nvary2(theta, wl, 1.0 - r_e * sld, d, sigma)

    # TODO: Fix corrections
    # FootprintCorrections
    # foocor = footprintcorr(Q, instrument)
    # Resolution corrections
    # R = resolutioncorr(R, TwoThetaQz, foocor, instrument, weight)

    return R * instrument.I0 + instrument.Ibkg


def OffSpecular(TwoThetaQz, ThetaQx, sample: Sample, instrument: Instrument):
    """Function that simulates the off-specular signal (not implemented)

    # BEGIN Parameters
    TwoThetaQz 1.0
    ThetaQx data.x
    # END Parameters
    """
    raise NotImplementedError("Not implemented use model interdiff insteads")


def SLD_calculations(z, item, sample: Sample, inst: Instrument):
    """Calculates the scatteringlength density as at the positions z
    if item is None or "all" the function returns a dictonary of values.
    Otherwise, it returns the item as identified by its string.

    # BEGIN Parameters
    z data.x
    item 'Re'
    # END Parameters
    """
    parameters: LayerParameters = sample.resolveLayerParameters()
    if hasattr(sample, "crop_sld") and sample.crop_sld != 0:
        crop_top_bottom = abs(sample.crop_sld)
        inter = Layer()
        if sample.crop_sld > 0:
            inter.d = sum(parameters.d[crop_top_bottom:-crop_top_bottom])
        else:
            inter.d = 5.0
            inter.dens = 0.1
            inter.b = 12.0 + 0j
            inter.f = 100.0 + 0j
        if len(parameters.dens) > 2 * crop_top_bottom:
            for fi in fields(Layer):
                key = fi.name
                value = getattr(parameters, key)
                val_start = value[:crop_top_bottom]
                val_end = value[-crop_top_bottom:]
                setattr(parameters, key, val_start + [getattr(inter, key)] + val_end)
    dens = array(parameters.dens, dtype=float32)
    # f = array(parameters['f'], dtype = complex64)
    e = AA_to_eV / inst.wavelength
    f = refl.cast_to_array(parameters.f, e).astype(complex64)
    b = array(parameters.b, dtype=complex64) * 1e-5
    abs_xs = array(parameters.xs_ai, dtype=float32) * 1e-4**2
    wl = inst.wavelength
    ptype = inst.probe
    magnetic = False
    mag_sld = 0
    sld_unit = "r_{e}/\\AA^{3}"
    if ptype == Probe.xray:
        sld = dens * f
    elif ptype in [Probe.neutron, Probe.ntof]:
        sld = dens * (wl**2 / 2 / pi * b - 1.0j * abs_xs * wl / 4 / pi) / 1e-6 / (wl**2 / 2 / pi)
        sld_unit = r"10^{-6}\AA^{-2}"
    else:
        magnetic = True
        sld = dens * (wl**2 / 2 / pi * b - 1.0j * abs_xs * wl / 4 / pi) / 1e-6 / (wl**2 / 2 / pi)
        magn = array(parameters.magn, dtype=float64)
        # Transform to radians
        magn_ang = array(parameters.magn_ang, dtype=float64) * pi / 180.0
        mag_sld = 2.645 * magn * dens * 10.0
        mag_sld_x = mag_sld * cos(magn_ang)
        mag_sld_y = mag_sld * sin(magn_ang)
        sld_unit = r"10^{-6}\AA^{-2}"

    d = array(parameters.d, dtype=float64)
    d = d[1:-1]
    # Include one extra element - the zero pos (substrate/film interface)
    int_pos = cumsum(r_[0, d])
    sigma = array(parameters.sigma, dtype=float64)[:-1] + 1e-7
    if z is None:
        z = arange(-sigma[0] * 5, int_pos.max() + sigma[-1] * 5, 0.5)
    if not magnetic:
        rho = sum((sld[:-1] - sld[1:]) * (0.5 - 0.5 * erf((z[:, newaxis] - int_pos) / sqrt(2.0) / sigma)), 1) + sld[-1]
        dic = {"Re": real(rho), "Im": imag(rho), "z": z, "SLD unit": sld_unit}
    else:
        sld_p = sld + mag_sld
        sld_m = sld - mag_sld
        rho_p = (
            sum((sld_p[:-1] - sld_p[1:]) * (0.5 - 0.5 * erf((z[:, newaxis] - int_pos) / sqrt(2.0) / sigma)), 1)
            + sld_p[-1]
        )
        rho_m = (
            sum((sld_m[:-1] - sld_m[1:]) * (0.5 - 0.5 * erf((z[:, newaxis] - int_pos) / sqrt(2.0) / sigma)), 1)
            + sld_m[-1]
        )
        rho_nucl = (rho_p + rho_m) / 2.0
        if inst.probe == Probe.npolsf and (magn_ang != 0.0).any():
            rho_mag_x = (
                sum(
                    (mag_sld_x[:-1] - mag_sld_x[1:]) * (0.5 - 0.5 * erf((z[:, newaxis] - int_pos) / sqrt(2.0) / sigma)),
                    1,
                )
                + mag_sld_x[-1]
            )
            rho_mag_y = (
                sum(
                    (mag_sld_y[:-1] - mag_sld_y[1:]) * (0.5 - 0.5 * erf((z[:, newaxis] - int_pos) / sqrt(2.0) / sigma)),
                    1,
                )
                + mag_sld_y[-1]
            )
            dic = {
                "Re non-mag": real(rho_nucl),
                "Im non-mag": imag(rho_nucl),
                "mag": real(rho_p - rho_m) / 2,
                "z": z,
                "mag_x": rho_mag_x,
                "mag_y": rho_mag_y,
                "SLD unit": sld_unit,
            }
        else:
            dic = {
                "Re non-mag": real(rho_nucl),
                "Im non-mag": imag(rho_nucl),
                "mag": real(rho_p - rho_m) / 2,
                "z": z,
                "SLD unit": sld_unit,
            }
    if item is None or item == "all":
        dic['Mass Density'] = MassDensity(z, item, sample)
        return dic
    else:
        try:
            return dic[item]
        except:
            raise ValueError("The chosen item, %s, does not exist" % item)

def MassDensity(z, item, sample: Sample):
    """
    Assuming all f-values are defined via Elements, calculate elemental and total mass density.
    """
    from .utils import __w_dict__

    parameters = sample.resolveLayerParameters()
    dens = array(parameters.dens, dtype=float32)
    mdens_by_element = {}
    for i, di in enumerate(dens):
        for number, element in getattr(parameters.f[i], 'composition', getattr(parameters.b[i], 'composition', ())):
            if element not in mdens_by_element:
                mdens_by_element[element] = [0. for j in range(i)]
            mdens_by_element[element].append(number*dens[i]*__w_dict__[element]*1.66053907)
        for element in mdens_by_element:
            if len(mdens_by_element[element])<=i:
                mdens_by_element[element].append(0.)

    d = array(parameters.d, dtype=float64)
    d = d[1:-1]
    int_pos = cumsum(r_[0, d])
    sigma = array(parameters.sigma, dtype=float64)[:-1] + 1e-7
    if z is None:
        z = arange(-sigma[0] * 5, int_pos.max() + sigma[-1] * 5, 0.5)

    dic = {'z': z, 'SLD unit': 'g/cm³', 'Sum': 0.*z}
    for element in mdens_by_element:
        mdele = array(mdens_by_element[element])
        rho_ele = sum((mdele[:-1] - mdele[1:]) * (0.5 - 0.5 * erf((z[:, newaxis] - int_pos) / sqrt(2.0) / sigma)), 1) + mdele[-1]
        dic[element.capitalize()] = rho_ele
        dic['Sum']+=dic[element.capitalize()]
    if item is None or item == "all":
        return dic
    else:
        try:
            return dic[item]
        except:
            raise ValueError("The chosen item, %s, does not exist" % item)

POL_CHANNELS = [Polarization.up_up, Polarization.up_down, Polarization.down_up, Polarization.down_down]


def PolSpecular(TwoThetaQz, p1, p2, F1, F2, sample: Sample, instrument: Instrument):
    """
    Specular reflectivity of polarized measurement with finite polarization.
    The polarization parameters are in accordance to the definition used in
    A.R. Wildes publication, Review of Scientific Instruments 70, 11 (1999)
    https://doi.org/10.1063/1.1150060

    pol_params = (p1, p2, f1, f2)
    p1: polarizer efficiency, 0: 100% spin-up, 1: 100% spin-down, 0.5: unpolarized
    p2: analyzer efficiency,  0: 100% spin-up, 1: 100% spin-down, 0.5: unpolairzed
    F1/F2: Flipper efficienty, 0: 100% efficient, 1: no flipping

    # BEGIN Parameters
    TwoThetaQz data.x
    p1 0.
    p2 0.
    F1 0.
    F2 0.
    # END Parameters

    """
    inst_pol = instrument.pol
    if not inst_pol in POL_CHANNELS:
        raise ValueError(f"Instrument polarization as to be one of {POL_CHANNELS}.")
    if instrument.probe != Probe.npolsf:
        raise ValueError("Polarization corrected simulation requires probe to be 'neutron pol spin flip'")

    instrument.pol = Polarization("uu")
    uu = Specular(TwoThetaQz, sample, instrument)
    instrument.pol = Polarization("dd")
    dd = Specular(TwoThetaQz, sample, instrument)
    instrument.pol = Polarization("ud")
    ud = Specular(TwoThetaQz, sample, instrument)
    du = ud
    instrument.pol = inst_pol

    P = get_pol_matrix(p1, p2, F1, F2)
    Pline = P[POL_CHANNELS.index(instrument.pol)]
    I = Pline[:, newaxis] * np.vstack([uu, ud, du, dd])
    return I.sum(axis=0)


SimulationFunctions = {
    "Specular": Specular,
    "PolSpecular": PolSpecular,
    "SpecularField": SpecularField,
    "OffSpecular": OffSpecular,
    "SLD": SLD_calculations,
    "EnergySpecular": EnergySpecular,
}

Sample.setSimulationFunctions(SimulationFunctions)


class TestSpecNX(ModelTestCase):
    # TODO: currently this only checks for raise conditions in the code above, check of results should be added

    def test_spec_xray(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, f=3e-5 + 1e-7j, dens=0.1)])],
            Ambient=Layer(),
            Substrate=Layer(f=5e-5 + 2e-7j, dens=0.1),
        )
        instrument = Instrument(
            probe=Probe.xray,
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=1.54,
        )
        with self.subTest("x-ray tth"):
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray q"):
            instrument.coords = Coords.q
            Specular(self.qz, sample, instrument)

        # resolution corrections
        with self.subTest("x-ray q-res-fast"):
            instrument.restype = ResType.fast_conv
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-fast-rel"):
            instrument.restype = ResType.fast_conv_rel
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-fast-var"):
            instrument.restype = ResType.fast_conv_var
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-full"):
            instrument.restype = ResType.full_conv_abs
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-full-rel"):
            instrument.restype = ResType.full_conv_rel
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-gauss"):
            instrument.restype = resolution_module.GaussianResolution(sigma=0.001)
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-trapezoid"):
            instrument.restype = resolution_module.TrapezoidResolution(inner_width=0.001, outer_width=0.002)
            instrument.restype.get_weight_example()
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-wrong"):
            instrument.restype = 123
            with self.assertRaises(ValueError):
                Specular(self.qz, sample, instrument)
        instrument.restype = ResType.none

        # footprint corrections
        with self.subTest("x-ray q-footprint-square"):
            instrument.footype = FootType.square
            Specular(self.qz, sample, instrument)
        instrument.coords = Coords.tth
        with self.subTest("x-ray tth-footprint-square"):
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-gauss"):
            instrument.footype = FootType.gauss
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-gauss-offset"):
            instrument.footype = GaussianBeamOffset(sigma=0.5, offset=0.1)
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-square-offset"):
            instrument.footype = SquareBeamOffset(width=0.1, offset=0.05)
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-trapezoid"):
            instrument.footype = TrapezoidBeam(inner_width=0.1, outer_width=0.2)
            Specular(self.tth, sample, instrument)
            instrument.footype = TrapezoidBeam(inner_width=0.1, outer_width=0.1)
            Specular(self.tth, sample, instrument)
            instrument.footype = TrapezoidBeam(inner_width=0.0, outer_width=0.2)
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-wrong"):
            instrument.footype = 123
            with self.assertRaises(ValueError):
                Specular(self.qz, sample, instrument)

        instrument.footype = FootType.none
        with self.subTest("x-ray tth-field"):
            SpecularField(self.tth, sample, instrument)

    def test_spec_neutron(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, b=3e-6, dens=0.1, magn=0.1, magn_ang=24.0)])],
            Ambient=Layer(b=1e-7, dens=0.1),
            Substrate=Layer(b=4e-6, dens=0.1),
        )
        instrument = Instrument(
            probe=Probe.neutron,
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=4.5,
            incangle=0.5,
        )
        with self.subTest("neutron tth"):
            Specular(self.tth, sample, instrument)
        with self.subTest("neutron q"):
            instrument.coords = Coords.q
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tof q-footprint"):
            instrument.probe = Probe.ntof
            instrument.footype = FootType.square
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tof q-footprint"):
            instrument.footype = FootType.none
            instrument.restype = ResType.full_conv_rel
            Specular(self.qz, sample, instrument)
            instrument.incangle = ones_like(self.qz) * 0.5
            Specular(self.qz, sample, instrument)
        instrument.incangle = 0.5
        instrument.restype = ResType.none
        with self.subTest("neutron-pol++ q"):
            instrument.probe = Probe.npol
            instrument.pol = Polarization.up_up
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-pol-- q"):
            instrument.pol = Polarization.down_down
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-polsa q"):
            instrument.pol = Polarization.asymmetry
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tofpol++ q"):
            instrument.probe = Probe.ntofpol
            instrument.pol = Polarization.up_up
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tofpol-- q"):
            instrument.pol = Polarization.down_down
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tofpolsa q"):
            instrument.pol = Polarization.asymmetry
            Specular(self.qz, sample, instrument)

        with self.subTest("neutron tthoffset"):
            instrument.probe = Probe.neutron
            instrument.tthoff = 0.1
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tof tthoffset"):
            instrument.probe = Probe.ntof
            instrument.tthoff = 0.1
            Specular(self.qz, sample, instrument)

        with self.subTest("neutron-sf++ q"):
            instrument.probe = Probe.npolsf
            instrument.pol = Polarization.up_up
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-sf-- q"):
            instrument.pol = Polarization.down_down
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-sf+- q"):
            instrument.pol = Polarization.up_down
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-sf-+ q"):
            instrument.pol = Polarization.down_up
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-sfsa q"):
            instrument.pol = Polarization.asymmetry
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-polspec q"):
            instrument.pol = Polarization.up_up
            PolSpecular(self.qz, 0.01, 0.01, 0.01, 0.01, sample, instrument)

    def test_energy(self):
        from .utils import fp

        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, b=3e-6, f=fp.Fe, dens=0.1, magn=0.1, magn_ang=24.0)])],
            Ambient=Layer(b=1e-7, dens=0.1),
            Substrate=Layer(b=4e-6, f=fp.Si, dens=0.1),
        )
        instrument = Instrument(
            probe=Probe.xray,
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=4.5,
            incangle=0.5,
        )
        energy = linspace(8000.0, 5000.0, 20)

        with self.subTest("energy xray"):
            EnergySpecular(energy, 1.5, sample, instrument)
        with self.subTest("energy xray q"):
            instrument.coords = Coords.q
            EnergySpecular(energy, 1.5, sample, instrument)

    def test_sld(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, f=2e-5 + 1e-7j, b=3e-6, dens=0.1, magn=0.1, magn_ang=24.0)])],
            Ambient=Layer(b=1e-7, dens=0.1),
            Substrate=Layer(b=4e-6, dens=0.1),
        )
        instrument = Instrument(
            probe=Probe.xray,
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=4.5,
            incangle=0.5,
        )
        with self.subTest("sld xray"):
            SLD_calculations(None, None, sample, instrument)
        with self.subTest("sld neutron"):
            instrument.probe = Probe.neutron
            SLD_calculations(None, None, sample, instrument)
        with self.subTest("sld neutron pol"):
            instrument.probe = Probe.npolsf
            SLD_calculations(None, None, sample, instrument)
        with self.subTest("sld neutron pol2"):
            sample.Stacks[0].Layers[0].magn_ang = 0.0
            SLD_calculations(None, None, sample, instrument)
        with self.subTest("sld neutron crop"):
            instrument.probe = Probe.neutron
            sample.crop_sld = 15
            sample.Stacks[0].Repetitions = 100
            SLD_calculations(None, None, sample, instrument)
            sample.crop_sld = -15
            sample.Stacks[0].Repetitions = 100
            SLD_calculations(None, None, sample, instrument)

    def test_inversion(self):
        sample = Sample(
            Stacks=[
                Stack(
                    Layers=[
                        Layer(d=50, sigma=4.0, b=3e-6, dens=0.1, magn=0.1, magn_ang=24.0),
                        Layer(d=100, sigma=3.0, b=3e-6, dens=0.1, magn=0.1, magn_ang=24.0),
                        Layer(d=150, sigma=2.0, b=3e-6, dens=0.1, magn=0.1, magn_ang=24.0),
                    ]
                )
            ],
            Ambient=Layer(b=1e-7, dens=0.1),
            Substrate=Layer(b=4e-6, dens=0.1, sigma=1.0),
        )
        sample_inv = -Sample(
            Stacks=[
                Stack(
                    Layers=[
                        Layer(d=150, sigma=3.0, b=3e-6, dens=0.1, magn=0.1, magn_ang=24.0),
                        Layer(d=100, sigma=4.0, b=3e-6, dens=0.1, magn=0.1, magn_ang=24.0),
                        Layer(d=50, sigma=1.0, b=3e-6, dens=0.1, magn=0.1, magn_ang=24.0),
                    ]
                )
            ],
            Ambient=Layer(b=4e-6, dens=0.1),
            Substrate=Layer(b=1e-7, dens=0.1, sigma=2.0),
        )
        instrument = Instrument(
            probe=Probe.neutron,
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=4.5,
            incangle=0.5,
        )
        ref = Specular(self.tth, sample, instrument)
        inv = Specular(self.tth, sample_inv, instrument)
        np.testing.assert_array_equal(ref, inv)


def standard_xray():
    """
    return the defied standard x-ray reflectivity to compare against other models
    """
    qz = linspace(0.01, 0.3, 15)
    return Specular(
        qz,
        Sample(
            Ambient=Layer(),
            Substrate=Layer(d=150.0, f=1e-5 + 1e-8j, dens=0.1),
            Stacks=[Stack(Layers=[Layer(d=150.0, f=2e-5 + 2e-8j, dens=0.1)])],
        ),
        Instrument(probe=Probe.xray, coords=Coords.q, wavelength=1.54, footype=FootType.none, restype=ResType.none),
    )



# run performance tests on python, numba and cuda versions of code

from multiprocessing import cpu_count
from numpy import pi, array, complex128, linspace, nan, hstack, log2
from timeit import Timer
from genx.models import lib


lib.USE_NUMBA = False

import platform


print(platform.processor())
import numba

from genx.models.lib.paratt import ReflQ as py_ReflQ
from genx.models.lib.paratt_numba import ReflQ as nb_ReflQ
from genx.models.lib.neutron_refl import Refl as py_ReflSF
from genx.models.lib.neutron_numba import Refl as nb_ReflSF


try:
    import numba.cuda


    numba.cuda.detect()  # display the cuda device used, in case something whent wrong and simulator is on
except:
    HAS_CUDA = False
else:
    from genx.models.lib.paratt_cuda import ReflQ as nc_ReflQ
    from genx.models.lib.neutron_cuda import Refl as nc_ReflSF
    HAS_CUDA = True

LAMBDA = 4.5

# SLD with wavlength factor included
sld_Si = 2.1e-06*LAMBDA**2/2./pi
sld_Fe = 8e-6*LAMBDA**2/2./pi
sld_Fe_p = 12.9e-6*LAMBDA**2/2./pi
sld_Fe_m = 2.9e-6*LAMBDA**2/2./pi
sld_Pt = 6.22e-6*LAMBDA**2/2./pi


def pot(sld):
    return (2*pi/LAMBDA)**2*(1.-(1.-sld)**2)


###### Models for Parratt evaluation ######
SMALL_REF = [
    linspace(0.001, 0.2, 100),  # Q
    LAMBDA,
    array([1.-sld_Si]+[1.-sld_Pt, 1.-sld_Fe_p]*1+[1.], dtype=complex128),  # n
    array([3.]+[100, 50]*1+[3.]),  # d
    array([10., ]*4),  # sigma
    ]

MED_REF = [
    linspace(0.001, 0.2, 400),  # Q
    LAMBDA,
    array([1.-sld_Si]+[1.-sld_Pt, 1.-sld_Fe_p]*10+[1.], dtype=complex128),  # n
    array([3.]+[100, 50]*10+[3.]),  # d
    array([10., ]*22),  # sigma
    ]

LQ_REF = [
    linspace(0.001, 0.2, 4000),  # Q
    LAMBDA,
    array([1.-sld_Si]+[1.-sld_Pt, 1.-sld_Fe_p]*10+[1.], dtype=complex128),  # n
    array([3.]+[100, 50]*10+[3.]),  # d
    array([10., ]*22),  # sigma
    ]

LS_REF = [
    linspace(0.001, 0.2, 400),  # Q
    LAMBDA,
    array([1.-sld_Si]+[1.-sld_Pt, 1.-sld_Fe_p]*100+[1.], dtype=complex128),  # n
    array([3.]+[100, 50]*100+[3.]),  # d
    array([10., ]*202),  # sigma
    ]

LARGE_REF = [
    linspace(0.001, 0.2, 4000),  # Q
    LAMBDA,
    array([1.-sld_Si]+[1.-sld_Pt, 1.-sld_Fe_p]*100+[1.], dtype=complex128),  # n
    array([3.]+[100, 50]*100+[3.]),  # d
    array([10., ]*202),  # sigma
    ]

###### Models for Matrix evaluation ######
SMALL_SF = [
    linspace(0.001, 0.2, 100),  # Q
    array([pot(sld_Si)]+[pot(sld_Pt), pot(sld_Fe_p)]*1+[0], dtype=complex128),  # Vp
    array([pot(sld_Si)]+[pot(sld_Pt), pot(sld_Fe_m)]*1+[0], dtype=complex128),  # Vm
    array([3.]+[100, 50]*1+[3.]),  # d
    array([0.0]+[0.0, 45*pi/180]*1+[0]),  # Mag_ang
    array([10., ]*4),  # sigma
    ]

MED_SF = [
    linspace(0.001, 0.2, 400),  # Q
    array([pot(sld_Si)]+[pot(sld_Pt), pot(sld_Fe_p)]*10+[0], dtype=complex128),  # Vp
    array([pot(sld_Si)]+[pot(sld_Pt), pot(sld_Fe_m)]*10+[0], dtype=complex128),  # Vm
    array([3.]+[100, 50]*10+[3.]),  # d
    array([0.0]+[0.0, 45*pi/180]*10+[0]),  # Mag_ang
    array([10., ]*22),  # sigma
    ]

LQ_SF = [
    linspace(0.001, 0.2, 4000),  # Q
    array([pot(sld_Si)]+[pot(sld_Pt), pot(sld_Fe_p)]*10+[0], dtype=complex128),  # Vp
    array([pot(sld_Si)]+[pot(sld_Pt), pot(sld_Fe_m)]*10+[0], dtype=complex128),  # Vm
    array([3.]+[100, 50]*10+[3.]),  # d
    array([0.0]+[0.0, 45*pi/180]*10+[0]),  # Mag_ang
    array([10., ]*22),  # sigma
    ]

LS_SF = [
    linspace(0.001, 0.2, 400),  # Q
    array([pot(sld_Si)]+[pot(sld_Pt), pot(sld_Fe_p)]*100+[0], dtype=complex128),  # Vp
    array([pot(sld_Si)]+[pot(sld_Pt), pot(sld_Fe_m)]*100+[0], dtype=complex128),  # Vm
    array([3.]+[100, 50]*100+[3.]),  # d
    array([0.0]+[0.0, 45*pi/180, 0., 90*pi/180.]*50+[0]),  # Mag_ang
    array([10., ]*202),  # sigma
    ]

LARGE_SF = [
    hstack([linspace(0.001, 0.2, 400), ]*10),  # Q - too fine steps breaks python variant
    array([pot(sld_Si)]+[pot(sld_Pt), pot(sld_Fe_p)]*100+[0], dtype=complex128),  # Vp
    array([pot(sld_Si)]+[pot(sld_Pt), pot(sld_Fe_m)]*100+[0], dtype=complex128),  # Vm
    array([3.]+[100, 50]*100+[3.]),  # d
    array([0.0]+[0.0, 45*pi/180, 0., 90*pi/180.]*50+[0]),  # Mag_ang
    array([10., ]*202),  # sigma
    ]


def time_call(call, init, repeat=1):
    # time a function all repeated times, goal is 1s run time
    T = Timer(call, init)
    t1 = T.timeit(repeat)
    rep = max(1, int(1./t1/5.))
    return min(T.repeat(5, rep))/rep


REF_ITMS = ['SMALL_REF', 'MED_REF', 'LQ_REF', 'LS_REF', 'LARGE_REF']
SF_ITMS = ['SMALL_SF', 'MED_SF', 'LQ_SF', 'LS_SF', 'LARGE_SF']

if hasattr(numba, 'set_num_threads'):
    print('Multithread scaling large model:')
    mthread_LS = []
    print(f'{"Cores":8s} | {"Parratt [ms]":16s} | {"[ms*cores]":12s} | '
          f'{"Matrix [ms]":16s} | {"[ms*cores]":12s}')
    for lprocs in range(int(log2(cpu_count()))+1):
        procs = 2**lprocs
        numba.set_num_threads(procs)
        t_parratt = time_call('nb_ReflQ(*LS_REF)', 'from __main__ import nb_ReflQ, LS_REF',
                              repeat=procs)

        t_matrix = time_call('nb_ReflSF(*LS_SF)', 'from __main__ import nb_ReflSF, LS_SF',
                             repeat=procs)
        print(f'{procs:8} | {t_parratt*1000:16.3f} | {t_parratt*1000*procs:12.3f} | '
              f'{t_matrix*1000:16.3f} | {t_matrix*1000*procs:12.3f}')
        mthread_LS.append((procs, t_parratt, t_matrix))

if hasattr(numba, 'set_num_threads'):
    print('Multithread scaling large data:')
    mthread_LQ = []
    print(f'{"Cores":8s} | {"Parratt [ms]":16s} | {"[ms*cores]":12s} | '
          f'{"Matrix [ms]":16s} | {"[ms*cores]":12s}')
    for lprocs in range(int(log2(cpu_count()))+1):
        procs = 2**lprocs
        numba.set_num_threads(procs)
        t_parratt = time_call('nb_ReflQ(*LQ_REF)', 'from __main__ import nb_ReflQ, LQ_REF',
                              repeat=procs)

        t_matrix = time_call('nb_ReflSF(*LQ_SF)', 'from __main__ import nb_ReflSF, LQ_SF',
                             repeat=procs)
        print(f'{procs:8} | {t_parratt*1000:16.3f} | {t_parratt*1000*procs:12.3f} | '
              f'{t_matrix*1000:16.3f} | {t_matrix*1000*procs:12.3f}')
        mthread_LQ.append((procs, t_parratt, t_matrix))

ref_results = []
print('Parratt algorithm:')
print('%16s | %16s | %16s | %10s | %16s | %12s | %16s | %12s'%('Points/Layers', 'Numpy [ms]',
                                                               'Numba (1x) [ms]', 'Speedup',
                                                               'Numba (%ix) [ms]'%cpu_count(), 'Speedup (x1)',
                                                               'Cuda [ms]', 'Speedup (x1)'))
for name in REF_ITMS:
    mod = eval(name)
    mlabel = '%i/%i'%(mod[0].shape[0], mod[2].shape[0])

    t_py = time_call('py_ReflQ(*%s)'%name, 'from __main__ import py_ReflQ, %s'%name)

    try:
        numba.set_num_threads(1)
    except AttributeError:
        t_nb1 = nan
    else:
        t_nb1 = time_call('nb_ReflQ(*%s)'%name, 'from __main__ import nb_ReflQ, %s'%name)
        numba.set_num_threads(cpu_count())
    t_nb = time_call('nb_ReflQ(*%s)'%name, 'from __main__ import nb_ReflQ, %s'%name)

    if HAS_CUDA:
        t_nc = time_call('nc_ReflQ(*%s)'%name, 'from __main__ import nc_ReflQ, %s'%name)
    else:
        t_nc = nan

    ref_results.append([t_py, t_nb1, t_nb, t_nc])
    print(f'{mlabel:16s} | {t_py*1000:16.3f} | {t_nb1*1000:16.3f} | {t_py/t_nb1:10.1f} | '
          f'{t_nb*1000:16.3f} | {t_nb1/t_nb:12.1f} | {t_nc*1000:16.3f} | {t_nb1/t_nc:12.1f}')

sf_results = []
print('Matrix algorithm:')
print('%16s | %16s | %16s | %10s | %16s | %12s | %16s | %12s'%('Points/Layers', 'Numpy [ms]',
                                                               'Numba (1x) [ms]', 'Speedup',
                                                               'Numba (%ix) [ms]'%cpu_count(), 'Speedup (x1)',
                                                               'Cuda [ms]', 'Speedup (x1)'))
for name in SF_ITMS:
    mod = eval(name)
    mlabel = '%i/%i'%(mod[0].shape[0], mod[2].shape[0])

    t_py = time_call('py_ReflSF(*%s)'%name, 'from __main__ import py_ReflSF, %s'%name)

    try:
        numba.set_num_threads(1)
    except AttributeError:
        t_nb1 = nan
    else:
        t_nb1 = time_call('nb_ReflSF(*%s)'%name, 'from __main__ import nb_ReflSF, %s'%name)
        numba.set_num_threads(cpu_count())
    t_nb = time_call('nb_ReflSF(*%s)'%name, 'from __main__ import nb_ReflSF, %s'%name)

    if HAS_CUDA:
        t_nc = time_call('nc_ReflSF(*%s)'%name, 'from __main__ import nc_ReflSF, %s'%name)
    else:
        t_nc = nan

    sf_results.append([t_py, t_nb1, t_nb, t_nc])
    print(f'{mlabel:16s} | {t_py*1000:16.3f} | {t_nb1*1000:16.3f} | {t_py/t_nb1:10.1f} | '
          f'{t_nb*1000:16.3f} | {t_nb1/t_nb:12.1f} | {t_nc*1000:16.3f} | {t_nb1/t_nc:12.1f}')

print("\n\nraw data:")
for (procs, t_pLS, t_mLS), (_, t_pLQ, t_mLQ) in zip(mthread_LS, mthread_LQ):
    print(f'Cores {procs}\t{t_pLS*1000}\t{t_pLQ*1000}\t'
          f'{t_mLS*1000}\t{t_mLQ*1000}')
for i, name in enumerate(REF_ITMS):
    mod = eval(name)
    nrep = mod[0]
    mlabel = '%i/%i'%(mod[0].shape[0], mod[2].shape[0])

    t_py, t_nb1, t_nb, t_nc = ref_results[i]
    print('Parratt %s\t%g\t%g\t%g\t%g\t%g\t%g\t%g'%(
        mlabel, t_py*1000,
        t_nb1*1000, t_py/t_nb1, t_nb*1000, t_nb1/t_nb,
        t_nc*1000, t_nb1/t_nc))
for i, name in enumerate(SF_ITMS):
    mod = eval(name)
    nrep = mod[0]
    mlabel = '%i/%i'%(mod[0].shape[0], mod[2].shape[0])

    t_py, t_nb1, t_nb, t_nc = sf_results[i]
    print('Matrix %s\t%g\t%g\t%g\t%g\t%g\t%g\t%g'%(
        mlabel, t_py*1000,
        t_nb1*1000, t_py/t_nb1, t_nb*1000, t_nb1/t_nb,
        t_nc*1000, t_nb1/t_nc))



'''Making sxrd data for simulations
P4mm symmetry k<h
'''
from numpy import c_, r_, array, arange, ones, savetxt

outfile = 'rods.dat'

h_min = 0
h_max = 3
k_min = 0
k_max = 3
l_min = 0.1
l_max = 3.5
l_step = 0.05

l = arange(l_min, l_max + l_step, l_step)

rods = array([[0,0,0,0,0]])[0:0]

for h in range(h_min, h_max + 1):
    for k in range(k_min, h+1):
        rods = r_[rods, c_[h*ones(l.shape), k*ones(l.shape), l, l*0, l*0]]

savetxt(outfile, rods)



"""
Test of data loader plugins.
"""

import unittest

from pathlib import Path
from os import sep

from genx.data import DataSet
from genx.plugins.utils import PluginHandler

DATA_DIR = Path(__file__).absolute().parent/"data"

class Dummy:
    def SetStatusText(self, message):
        pass

class TestDataLoaders(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        head = DATA_DIR.parent.parent
        cls._handler = PluginHandler(None, str(head / "genx" / "plugins")+sep, "data_loaders")
        cls.data_loaders = {}
        parent = Dummy()
        for dl in cls._handler.get_possible_plugins():
            try:
                cls._handler.load_plugin(dl)
                cls._handler.loaded_plugins[dl].parent = parent
                cls.data_loaders[dl] = cls._handler.loaded_plugins[dl]
            except Exception as error:
                print(error)

    def test_d17_legacy(self):
        data = DataSet()
        self.data_loaders['d17_legacy'].LoadData(data, DATA_DIR / "D17_SiO.out")

    def test_default(self):
        data = DataSet()
        self.data_loaders['default'].LoadData(data,DATA_DIR / "xray-tutorial.dat")

    def test_auto(self):
        # use all wildcards provided by the auto loaded to try to open files in the data folder
        files = []
        auto =self.data_loaders['auto']
        for wc in auto.wildcard.split(';'):
            files += list(DATA_DIR.glob(wc))
        for i,fi in enumerate(set(files)):
            data = DataSet()
            auto.LoadData(data, str(fi))
            with self.subTest(msg=f'{fi.name} by {auto._last_loader_used}'):
                self.assertTrue(len(data.x_raw)>0)
                self.assertEqual(len(data.x_raw), len(data.y_raw))
                self.assertEqual(len(data.x_raw), len(data.error_raw))
                self.assertEqual(len(data.extra_data_raw.keys()), len(data.extra_commands.keys()))
                self.assertEqual(len(data.extra_data_raw.keys()), len(data.extra_data.keys()))



"""
Script to download the FFAST library and convert it to nff files
"""

__author__ = "Matts Bjorck"

import os
import time
import urllib.error
import urllib.parse
import urllib.request

import numpy


def get_html_code(Z):
    adress = "http://physics.nist.gov/cgi-bin/ffast/ffast.pl?Z=%d&Formula=&gtype=4&lower=&upper=&density=&frames=no" % Z
    f = urllib.request.urlopen(adress)
    return f.readlines()


def find_line(lines, string_to_find):
    for i, line in enumerate(lines):
        if line.find(string_to_find) != -1:
            return line, i
    return None, None


def parse_page(lines):
    # Find the element
    line, i = find_line(lines, "(Z =")
    element = line.split("<b>")[1].split("&#160;")[0].strip()
    print("Element: ", element)

    # Find the Relativistic correction estimate
    line, i = find_line(lines, "Relativistic correction estimate")
    rel_corr = float(line.split("=")[1].split(",")[1].split("<i>e</i>")[0].strip())
    print("Relativistic correction estimate 3/5CL: ", rel_corr)

    # Find the Nuclear Thomson correction
    line, i = find_line(lines, "Nuclear Thomson correction")
    nuc_corr = float(line.split("=")[1].split("<i>e</i>")[0].strip())
    print("Nuclear Thomson correction: ", nuc_corr)

    # Find the data
    line, i = find_line(lines, "<b>Form Factors,")
    data_lines = lines[i + 3 : -3]
    data = numpy.fromstring(" ".join(data_lines), sep=" ").reshape((len(data_lines), 8))
    table = numpy.c_[data[:, 0] * 1e3, data[:, 1] + rel_corr + nuc_corr, data[:, 2]]
    return table, element


def create_nff(Z, path):
    html_code = get_html_code(Z)
    table, element = parse_page(html_code)
    with open(os.path.join(path, "%s.nff" % element.lower()), "w") as f:
        f.write("E (eV)\tf1 e/atom\tf2 e/atom\n")
        numpy.savetxt(f, table)


def create_readme(path):
    readme_str = """README
    These tables were generated %s.
    The following files were created by the download_nist.py script located in the GenX distribution. The data
    were generated from the NIST X-ray Form Factor, Attenuation and Scattering Tables (FFAST) located at:
    http://www.nist.gov/pml/data/ffast/index.cfm.
    
    Each element's scattering factors were evaluated according to eq. 4 in
    C. T. Chandler J. Phys. Chem. Ref. Data 29(4) 597-1048 (2000). Thus the scattering factors were calculated
    according to
        f1 = f_1 + f_rel(3/5 CL) + f_NT,
        f2 = f_2.
    
    According to the homepage, see above, the tables should be referenced according to:
        Chantler, C.T., Olsen, K., Dragoset, R.A., Chang, J., Kishore, A.R., Kotochigova, S.A., and Zucker, D.S. (2005),
        X-Ray Form Factor, Attenuation and Scattering Tables (version 2.1).
        [Online] Available: http://physics.nist.gov/ffast [%s].
        National Institute of Standards and Technology, Gaithersburg, MD.
        Originally published as Chantler, C.T., J. Phys. Chem. Ref. Data 29(4), 597-1048 (2000);
        and Chantler, C.T., J. Phys. Chem. Ref. Data 24, 71-643 (1995).
            """ % (
        time.ctime(),
        time.ctime(),
    )
    with open(os.path.join(path, "README.txt"), "w") as f:
        f.write(readme_str)


if __name__ == "__main__":
    lib_path = "f1f2_nist"
    for z in range(1, 93):
        create_nff(z, lib_path)
    create_readme(lib_path)



'''
A utillity that exports all png files in the scripts directory
to a python file using img2py
'''

import wx.tools.img2py as impy
import os

files = [file for file in os.listdir('.') if file[-4:] == '.png']

appending = False

for file in files:
    impy.img2py(file, 'images.py', append = appending, imgName = file[:-4],
                functionCompatibile=True, functionCompatible=True)
    appending = True



import warnings

from dataclasses import dataclass
from logging import debug

import matplotlib
import matplotlib.axes
import wx
import numpy as np
import matplotlib.pyplot as plt

from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg
from matplotlib.figure import Figure
from numpy import (arange, arccos, arcsin, arctan, array, cos, exp, hstack, linspace, log, log10, pi, sin, sqrt, tan,
                   vstack)
from wx.py.editwindow import EditWindow

from ..core.config import BaseConfig, Configurable
from ..data import DataList, DataSet


# ==============================================================================
class SimplePlotPanel(wx.Panel):
    ax: matplotlib.axes.Axes

    def __init__(
        self, parent, id=-1, color=None, dpi=None, style=wx.NO_FULL_REPAINT_ON_RESIZE | wx.EXPAND | wx.ALL, **kwargs
    ):
        wx.Panel.__init__(self, parent, id=id, style=style, **kwargs)
        if dpi is None:
            dpi = wx.GetApp().dpi_scale_factor * 96.0  # wx.GetDisplayPPI()[0]
        self.parent = parent
        debug("init PlotPanel - setup figure")
        self.figure = Figure(figsize=(1.0, 1.0), dpi=dpi)
        debug("init PlotPanel - setup canvas")
        self.canvas = FigureCanvasWxAgg(self, -1, self.figure)
        self.SetColor(color)
        self._resizeflag = True
        self.print_size = (15.0 / 2.54, 12.0 / 2.54)
        # self._SetSize()

        # debug('init PlotPanel - bind events')
        self.Bind(wx.EVT_IDLE, self._onIdle)
        self.Bind(wx.EVT_SIZE, self._onSize)

        cursor = wx.Cursor(wx.CURSOR_CROSS)
        self.canvas.SetCursor(cursor)
        self.ax = self.figure.add_subplot(111)

        # Create the drawing bitmap
        self.bitmap = wx.Bitmap(1, 1, depth=wx.BITMAP_SCREEN_DEPTH)
        #        DEBUG_MSG("__init__() - bitmap w:%d h:%d" % (w,h), 2, self)
        self._isDrawn = False
        debug("end init PlotPanel")

    def clear(self):
        self.figure.clear()
        self.ax = self.figure.add_subplot(111)

    def SetColor(self, rgbtuple=None):
        """Set the figure and canvas color to be the same"""
        if not rgbtuple:
            rgbtuple = self.parent.GetBackgroundColour()
            # wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE).Get()
        col = [c / 255.0 for c in rgbtuple]
        self.figure.set_facecolor(col)
        self.figure.set_edgecolor(col)
        self.canvas.SetBackgroundColour(wx.Colour(*rgbtuple))

    def _onSize(self, evt):
        self._resizeflag = True
        self._SetSize()

    def _onIdle(self, evt):
        if self._resizeflag:
            self._resizeflag = False
            self._SetSize()

    def _SetSize(self):
        """This method can be called to force the Plot to be a desired
        size which defaults to the ClientSize of the Panel.
        """
        pixels = self.GetMinSize()

        self.canvas.SetSize(pixels)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            try:
                self.figure.tight_layout(h_pad=0)
            except ValueError:
                pass

    def flush_plot(self):
        # with warnings.catch_warnings():
        #     warnings.simplefilter("ignore", UserWarning)
        #     self.figure.tight_layout(h_pad=0)
        self.canvas.draw()


@dataclass
class PublicationConfig(BaseConfig):
    section = "publication graph"

    width: float = 3.39
    heigth: float = 2.36
    font_size: int = int(matplotlib.rcParams["font.size"])
    font_familty: int = wx.FONTFAMILY_DEFAULT
    font_face: str = matplotlib.rcParams["font.sans-serif"][0]

    start_text: str = repr(
        """## Create plots using matplotlib related to your model.
## Uncomment relevant lines to save/show different graphs.
## For other matplotlib rc style options see 
## https://matplotlib.org/stable/tutorials/introductory/customizing.html
fig.set_facecolor('white')
rcParams['font.family'] = 'sans-serif'
rcParams['font.size'] = font_size # dialog entry
rcParams['font.sans-serif'] = font_face # dialog entry
rcParams['font.weight'] = font_weight # dialog entry
rcParams['mathtext.fontset'] = 'dejavusans' # 'dejavuserif'|'cm'|'stix'|'stixsans'

#outimg = save_file_dialog('*.png') # select filename for export
############ Plot of Reflectivity curves ###############
clear()
for di in data:
    if not di.show:
        continue
    errorbar(di.x, di.y, yerr=di.error, label="data-"+di.name, **di.data_kwds, zorder=2)
    semilogy(di.x, di.y_sim, label="sim-"+di.name, **di.sim_kwds, zorder=5)
xlabel(model.__xlabel__)
ylabel(model.__ylabel__)
legend()
tight_layout(pad=0.5)
#savefig(outimg, dpi=300) # save to selected image file
show() # show in GUI graph

########### Plot of SLD profiles ##############
clear()
sld = SLD[0] # assuming all dataset represent the same model
for key, value in sld.items():
    if key in ['z', 'SLD unit', 'Mass Density']:
        continue
    plot(sld['z'], value, label=key)
xlabel("z [Å]")
ylabel(f"SLD [${sld['SLD unit']}$]")
legend()
tight_layout(pad=0.5)
#savefig(outimg[:-4]+'_sld.png', dpi=300) # save to image file with changed suffix
#show() # show in GUI graph

########### Plot of SLD profiles ##############
clear()
md = SLD[0]['Mass Density'] # assuming all dataset represent the same model
## replace standard line colors with colormap
## See https://matplotlib.org/stable/users/explain/colors/colormaps.html
axes.set_prop_cycle(color=cm.gist_rainbow(np.linspace(0,1,len(md.keys())-1)))
for key, value in md.items():
    if key in ['z', 'SLD unit']:
        continue
    plot(md['z'], value, label=key)
xlabel("z [Å]")
ylabel(f"density [${md['SLD unit']}$]")
legend()
tight_layout(pad=0.5)
#savefig(outimg[:-4]+'_dens.png', dpi=300) # save to image file with changed suffix
#show() # show in GUI graph

""".splitlines()
    )


class PublicationDialog(wx.Dialog, Configurable):
    opt: PublicationConfig

    def __init__(self, parent, id=-1, data: DataList = None, module=None, SLD: dict = None):
        wx.Dialog.__init__(
            self, parent, id=id, style=wx.MAXIMIZE_BOX | wx.MINIMIZE_BOX | wx.RESIZE_BORDER | wx.DEFAULT_DIALOG_STYLE
        )
        Configurable.__init__(self)
        self.ReadConfig()
        self.SetTitle("Custom plotting for Publication")

        self.data = data
        self.module = module
        self.SLD = SLD
        self.dpi_scale_factor = parent.dpi_scale_factor
        main_box = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(main_box)
        splitter = wx.SplitterWindow(self, wx.ID_ANY, style=wx.SP_3D | wx.SP_BORDER | wx.SP_LIVE_UPDATE)
        main_box.Add(splitter, 1, wx.EXPAND)

        left_panel = wx.Panel(splitter, wx.ID_ANY)
        vbox = wx.BoxSizer(wx.VERTICAL)
        left_panel.SetSizer(vbox)
        right_panel = wx.Panel(splitter, wx.ID_ANY)
        vbox2 = wx.BoxSizer(wx.VERTICAL)
        right_panel.SetSizer(vbox2)
        splitter.SplitVertically(left_panel, right_panel)
        splitter.SetSashPosition(550)
        splitter.SetSashGravity(0.66)

        # ------------ Left Panel items --------------
        self.tinput = EditWindow(left_panel, wx.ID_ANY)
        self.tinput.SetBackSpaceUnIndents(True)
        self.tinput.SetText("\n".join(eval(self.opt.start_text)))

        vbox.Add(self.tinput, 1, wx.EXPAND, 0)
        button = wx.Button(left_panel, label="plot")
        vbox.Add(button, 0, wx.EXPAND, 0)

        # ------------ Right Panel items --------------
        self.plot_area = wx.ScrolledWindow(right_panel)
        self.plot_area.SetScrollbars(1, 1, 1, 1)
        pavbox = wx.BoxSizer(wx.VERTICAL)
        self.plot_area.SetSizer(pavbox)
        self.plot_area.SetAutoLayout(True)
        vbox2.Add(self.plot_area, 1, wx.EXPAND, 0)
        self.plot = SimplePlotPanel(self.plot_area)
        self.plot.SetMinSize(wx.Size(300, 300))
        pavbox.Add(self.plot, 1, wx.EXPAND, 0)
        self.error_text = wx.StaticText(right_panel)
        self.error_text.SetBackgroundColour(wx.Colour(255, 150, 150))
        self.error_text.Hide()
        vbox2.Add(self.error_text, 0, wx.EXPAND, 0)
        hbox2 = wx.BoxSizer(wx.HORIZONTAL)
        vbox2.Add(hbox2, 0, wx.EXPAND, 0)
        self.width = wx.SpinCtrlDouble(
            right_panel, min=0.1, max=30.0, initial=self.opt.width, inc=0.01,
            style=wx.SP_ARROW_KEYS|wx.TE_PROCESS_ENTER
        )
        self.height = wx.SpinCtrlDouble(
            right_panel, min=0.1, max=30.0, initial=self.opt.heigth, inc=0.01,
            style=wx.SP_ARROW_KEYS|wx.TE_PROCESS_ENTER
        )
        hbox2.Add(wx.StaticText(right_panel, label="Width (inches)"))
        hbox2.Add(self.width, 0, wx.FIXED_MINSIZE | wx.RIGHT, 10)
        hbox2.Add(wx.StaticText(right_panel, label="Height (inches)"))
        hbox2.Add(self.height, 0, wx.FIXED_MINSIZE | wx.RIGHT, 10)
        hbox2 = wx.BoxSizer(wx.HORIZONTAL)
        vbox2.Add(hbox2, 0, wx.EXPAND, 0)
        self.font = wx.FontPickerCtrl(right_panel, style=wx.FNTP_FONTDESC_AS_LABEL)
        self.font.SetSelectedFont(
            wx.Font(
                self.opt.font_size,
                self.opt.font_familty,
                wx.FONTSTYLE_NORMAL,
                wx.FONTWEIGHT_NORMAL,
                False,
                faceName=self.opt.font_face,
            )
        )
        hbox2.Add(wx.StaticText(right_panel, label="Font"))
        hbox2.Add(self.font, 0, wx.EXPAND, 0)

        hbox2 = wx.BoxSizer(wx.HORIZONTAL)
        vbox2.Add(hbox2, 0, wx.EXPAND, 0)
        save_button = wx.Button(right_panel, label="store current script and settings as default")
        hbox2.Add(save_button, 0, wx.EXPAND, 0)

        self.Bind(wx.EVT_BUTTON, self.OnPlot, button)
        self.Bind(wx.EVT_BUTTON, self.OnStoreDefaults, save_button)
        self.Bind(wx.EVT_COMMAND_ENTER, self.OnPlot)
        self.Bind(wx.EVT_SPINCTRLDOUBLE, self.OnPlot)
        self.SetSize(wx.Size(800, 800))

    def SaveFromScript(self, wildcard="*.png"):
        # called from within the script to select file name to save to
        with wx.FileDialog(self, "Save Plot",
                           wildcard=f"Imag files ({wildcard})|{wildcard}",
                           style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT) as fileDialog:
            if fileDialog.ShowModal()==wx.ID_CANCEL:
                return  # the user changed their mind

            # save the current contents in the file
            return fileDialog.GetPath()

    def OnPlot(self, event):
        txt = self.tinput.GetText()
        font = self.font.GetSelectedFont()
        weights = {wx.FONTWEIGHT_BOLD: "bold", wx.FONTWEIGHT_NORMAL: "normal"}
        env = dict(
            rcParams=matplotlib.rcParams,
            data=self.data,
            model=self.module,
            SLD=self.SLD,
            show=self.plot.flush_plot,
            fig=self.plot.figure,
            canvas=self.plot.canvas,
            savefig=self.plot.figure.savefig,
            tight_layout=self.plot.figure.tight_layout,
            font_size=font.GetPointSize(),
            font_face=font.GetFaceName(),
            font_weight=weights[font.GetWeight()],
            save_file_dialog=self.SaveFromScript,
            np=np,
            cm=plt.cm,
        )

        def clear():
            self.plot.clear()
            ax = self.plot.ax
            env["plot"] = ax.plot
            env["semilogy"] = ax.semilogy
            env["errorbar"] = ax.errorbar
            env["title"] = ax.set_title

            env["xlabel"] = ax.set_xlabel
            env["ylabel"] = ax.set_ylabel
            env["legend"] = ax.legend
            env["axes"] = ax

        env["clear"] = clear

        w = int(self.width.GetValue() * 72 * self.dpi_scale_factor * 2)
        h = int(self.height.GetValue() * 72 * self.dpi_scale_factor * 2)
        self.plot.figure.dpi = int(72 * self.dpi_scale_factor * 2)
        size = wx.Size(w, h)
        self.plot.SetMinSize(size)
        self.plot.SetSize(size)
        self.plot_area.FitInside()
        clear()
        try:
            exec(txt, env)
        except Exception as e:
            self.error_text.SetLabel(e.__class__.__name__ + ": " + str(e))
            self.error_text.Show()
        else:
            self.error_text.Hide()

        # make sure we reset the default rc values
        matplotlib.rcdefaults()
        self.Layout()

    def OnStoreDefaults(self, event):
        self.opt.start_text = repr(self.tinput.GetText().splitlines())
        self.opt.width = self.width.GetValue()
        self.opt.heigth = self.height.GetValue()

        font = self.font.GetSelectedFont()
        self.opt.font_size = font.GetPointSize()
        self.opt.font_face = font.GetFaceName()
        self.opt.font_familty = font.GetFamily()
        self.WriteConfig(default=True)



"""
Library for specular and off-specular x-ray reflectivity
========================================================
interdiff is a model for specular and off specular simulations including
the effects of interdiffusion in the calculations. The specular
simulations is conducted with Parrats recursion formula. The
off-specular, diffuse calculations are done with the distorted wave Born
approximation (DWBA) as derived by Holy and with the extensions done by
Wormington to include diffuse interfaces.

Classes
-------

"""

from dataclasses import dataclass, field, fields
from typing import List

import numpy as np

# import all special footprint functions
from .lib import footprint as footprint_module
from .lib import neutron_refl as MatrixNeutron
from .lib import offspec
from .lib import paratt as Paratt
from .lib import refl_base as refl
from .lib import resolution as resolution_module
from .lib.base import AltStrEnum
from .lib.footprint import *
from .lib.instrument import *
from .lib.physical_constants import AA_to_eV, muB_to_SL, r_e
from .lib.resolution import *
from .lib.testing import ModelTestCase

# Preamble to define the parameters needed for the models outlined below:
ModelID = "MingInterdiff"

q_limit = 1e-10
""" Minimum allowed q-value """

__xlabel__ = "q [Å$^{-1}$]"
__ylabel__ = "Instnsity [a.u.]"


class Coords(AltStrEnum):
    q = "q"
    tth = "2θ"
    alternate_tth = "tth"


class ResType(AltStrEnum):
    none = "no conv"
    fast_conv = "fast conv"
    fast_conv_var = "fast conv + varying res."
    full_conv_abs = "full conv and varying res."


class FootType(AltStrEnum):
    none = "no corr"
    gauss = "gauss beam"
    square = "square beam"


@dataclass
class Layer(refl.ReflBase):
    """
    Representing a layer in the sample structur.

    ``d``
       The thickness of the layer in AA (Angstroms = 1e-10m)
    ``dens``
       The density of formula units in units per Angstroms. Note the units!
    ``sigma``
       The total root mean square *roughness* of the top interface of the layer in
       Angstroms. This includes both the extended "wavy" roughness sigmar and the atomic-scale
       interface broadening due to interdiffusion sigmai. It is defined as sqrt(sigmar**2+sigmai**2).
    ``sigmai``
       The root mean square *interdiffusion* of the top interface of the
       layer in Angstroms.
    ``f``
       The x-ray scattering length per formula unit in electrons. To be
       strict it is the number of Thompson scattering lengths for each
       formula unit.
    """

    d: float = 0.0
    dens: float = 1.0
    sigma: float = 5.0
    sigmai: float = 0.0
    f: complex = 1e-20j

    Units = {
        "d": "AA",
        "dens": "at./AA",
        "sigma": "AA",
        "sigmai": "AA",
        "f": "el./at.",
    }

    Groups = [("General", ["d", "dens", "sigma", "sigmai"]), ("X-Ray", ["f"])]

    DEFAULT_FIT_PARAMS = ["d", "dens", "sigma", "sigmai"]

@dataclass
class LayerParameters:
    d: List[float]
    dens: List[float]
    sigma: List[float]
    sigmai: List[float]
    f: List[complex]


@dataclass
class Stack(refl.StackBase):
    """
    A collection of Layer objects that can be repeated.

    ``Layers``
       A ``list`` consiting of ``Layers`` in the stack the first item is
       the layer closest to the bottom
    ``Repetitions``
       The number of repsetions of the stack
    """

    Layers: List[Layer] = field(default_factory=list)
    Repetitions: int = 1


@dataclass
class Sample(refl.SampleBase):
    """
    Describe global sample by listing ambient, substrate and layer parameters.

    ``Stacks``
       A ``list`` consiting of ``Stacks`` in the stacks the first item is
       the layer closest to the bottom
    ``Ambient``
       A ``Layer`` describing the Ambient (enviroment above the sample).
       Only the scattering lengths and density of the layer is used.
    ``Substrate``
       A ``Layer`` describing the substrate (enviroment below the sample).
       Only the scattering lengths, density and roughness of the layer is
       used.
    ``eta_z``
       The out-of plane (vertical) correlation length of the roughness in
       the sample. Given in AA.
    ``eta_x``
       The in-plane global correlation length (it is assumed equal for all
       layers). Given in AA.
    ``h``
       The jaggedness parameter, should be between 0 (jagged) and 1.0 (smooth). This describes
       how jagged the interfaces are. This is also a global parameter for
       all interfaces. (See e.g.: [book](https://doi.org/10.1007/978-1-4612-3784-6_2) )
    """

    Stacks: List[Stack] = field(default_factory=list)
    Ambient: Layer = field(default_factory=Layer)
    Substrate: Layer = field(default_factory=Layer)
    eta_z: float = 100.0
    eta_x: float = 100.0
    h: float = 1.0

    _layer_parameter_class = LayerParameters


@dataclass
class Instrument(refl.ReflBase):
    """
    Specify parameters of the probe and reflectometry instrument.

    ``wavelength``
        The wavelength of the radiation given in AA (Angstroms)
    ``coords``
        The coordinates of the data given to the SimSpecular function. The
        available alternatives are: 'q' or '2θ'. Alternatively the numbers 0 (q)
        or 1 (tth) can be used.
    ``I0``
        The incident intensity (a scaling factor)
    ``Ibkg``
        The background intensity. Added as a constant value to the calculated
        reflectivity
    ``restype``
        Describes the rype of the resolution calculated. One of the
        alterantives: 'no conv', 'fast conv', 'full conv and varying res.',
        'fast conv + varying res.', 'full conv and varying res. (dx/x)', 'fast
        conv + varying res. (dx/x)'. The respective numbers 0-3 also works. Note
        that fast convolution only alllows a single value into res wheras the
        other can also take an array with the same length as the x-data (varying
        resolution)
    ``res``
        The resolution of the instrument given in the coordinates of ``coords``.
        This assumes a gaussian resolution function and ``res`` is the standard
        deviation of that gaussian. If ``restype`` has (dx/x) in its name the
        gaussian standard deviation is given by res*x where x is either in tth
        or q.
    ``respoints``
        The number of points to include in the resolution calculation. This is
        only used for 'full conv and vaying res.' and 'fast conv + varying res'.
    ``resintrange``
        Number of standard deviatons to integrate the resolution function times
        the reflectivity over
    ``footype``
        Which type of footprint correction is to be applied to the simulation.
        One of: 'no corr', 'gauss beam' or 'square beam'. Alternatively, the
        number 0-2 are also valid. The different choices are self expnalatory.
    ``beamw``
        The width of the beam given in mm. For 'gauss beam' it should be the
        standard deviation. For 'square beam' it is the full width of the beam.
    ``samplelen``
        The length of the sample given in mm
    ``taylor_n``
        The number terms taken into account in the taylor expansion of the
        fourier integral of the correlation function. More terms more accurate
        calculation but also much slower.
    """

    wavelength: float = 1.54
    coords: Coords = "2θ"
    I0: float = 1.0
    I0_offspec: float = 1.0
    Ibkg: float = 0.0
    restype: ResType = "no conv"
    res: float = 0.001
    detres: float = 0.01
    respoints: int = 5
    resintrange: float = 2.0
    footype: FootType = "no corr"
    beamw: float = 0.01
    samplelen: float = 10.0
    taylor_n: int = 2

    Units = {
        "wavelength": "Å",
        "coords": "",
        "I0": "arb.",
        "I0_offspec": "arb.",
        "Ibkg": "arb.",
        "restype": "",
        "res": "[coord]",
        "detres": "[coord]",
        "respoints": "pts.",
        "resintrange": "[coord]",
        "footype": "",
        "beamw": "mm",
        "samplelen": "mm",
        "2θ": "°",
        "tth": "°",
        "q": "Å$^-1$",
    }

    Groups = [
        ("Radiation", ["wavelength", "I0", "I0_offspec", "Ibkg"]),
        ("Resolution", ["restype", "res", "detres", "respoints", "resintrange"]),
        ("X-Coordinates", ["coords"]),
        ("DWBA", ["taylor_n"]),
        ("Footprint", ["footype", "beamw", "samplelen"]),
    ]

    DEFAULT_FIT_PARAMS = ["I0", "Ibkg", "Beamw", "res"]


def Specular(TwoThetaQz, sample: Sample, instrument: Instrument):
    """Simulate the specular signal from sample when proped with instrument

    # BEGIN Parameters
    TwoThetaQz data.x
    # END Parameters
    """
    # preamble to get it working with my class interface
    global __xlabel__
    __xlabel__ = "q [Å$^{-1}$]"

    restype = instrument.restype
    weight = 0
    if isinstance(restype, resolution_module.Resolution):
        (TwoThetaQz, weight) = restype(TwoThetaQz[:], instrument.respoints, instrument.resintrange)
    elif restype == ResType.full_conv_abs:
        res_function = GaussianResolution(sigma=instrument.res)
        (TwoThetaQz, weight) = res_function(TwoThetaQz[:], instrument.respoints, instrument.resintrange)

    if instrument.coords == Coords.tth:
        theta = TwoThetaQz / 2
        __xlabel__ = "2θ [°]"
    elif instrument.coords == Coords.q:
        theta = arcsin(TwoThetaQz / 4 / pi * instrument.wavelength) * 180.0 / pi

    lamda = instrument.wavelength
    parameters: LayerParameters = sample.resolveLayerParameters()
    dens = array(parameters.dens, dtype=float64)
    # print [type(f) for f in parameters['f']]
    f = array(parameters.f, dtype=complex128)
    n = 1 - dens * r_e * lamda**2 / 2 / pi * f
    d = array(parameters.d, dtype=float64)

    sigma = array(parameters.sigma, dtype=float64)

    R = Paratt.Refl(theta, lamda, n, d, sigma) * instrument.I0

    # FootprintCorrections
    foocor = 1.0
    footype = instrument.footype
    beamw = instrument.beamw
    samlen = instrument.samplelen
    if footype == FootType.gauss:
        foocor = GaussIntensity(theta, samlen / 2.0, samlen / 2.0, beamw)
    elif footype == FootType.square:
        foocor = SquareIntensity(theta, samlen, beamw)
    elif footype == FootType.none:
        pass
    elif isinstance(footype, footprint_module.Footprint):
        foocor = footype(theta, samlen)
    else:
        raise ValueError("The choice of footprint correction, footype," "is WRONG")

    R = R[:] * foocor
    restype = instrument.restype
    if restype == ResType.fast_conv:
        R = ConvoluteFast(TwoThetaQz, R, instrument.res, range=instrument.resintrange)
    elif restype == ResType.full_conv_abs or isinstance(restype, resolution_module.Resolution):
        R = ConvoluteResolutionVector(TwoThetaQz, R, weight)
    elif restype == ResType.fast_conv_var:
        R = ConvoluteFastVar(TwoThetaQz, R, instrument.res, range=instrument.resintrange)
    elif restype == ResType.none:
        pass
    else:
        raise ValueError("The choice of resolution type, restype," "is WRONG")
    return R + instrument.Ibkg


def OffSpecularMingInterdiff(TwoThetaQz, ThetaQx, sample: Sample, instrument: Instrument):
    """Function that simulates the off-specular signal (not implemented)

    # BEGIN Parameters
    TwoThetaQz 1.0
    ThetaQx data.x
    # END Parameters
    """
    lamda = instrument.wavelength
    k = 2 * pi / lamda
    if instrument.coords == Coords.tth:
        alphaR1 = ThetaQx
        betaR1 = TwoThetaQz - ThetaQx
        if (betaR1 > 0).all() or (betaR1 < 0).all():
            longitudinal = True
        else:
            longitudinal = False
            cen_idx = where(betaR1 >= 0)[0][0]
        qx = k * (cos(alphaR1 * pi / 180) - cos(betaR1 * pi / 180))
        qz = k * (sin(alphaR1 * pi / 180) + sin(betaR1 * pi / 180))
        tth = TwoThetaQz
    else:
        qz = TwoThetaQz
        qx = ThetaQx
        if (qx > 0).all() or (qx < 0).all():
            longitudinal = True
        else:
            longitudinal = False
            cen_idx = where(qx >= 0)[0][0]
        tth = arcsin(qz / 4.0 / pi * lamda) * 2.0
        alphaR1 = (arctan2(qx, qz) + tth / 2.0) * 180.0 / pi
        betaR1 = tth - alphaR1

    parameters: LayerParameters = sample.resolveLayerParameters()

    def toarray(a, code):
        a = list(a)
        a.reverse()
        return array(a, dtype=code)

    dens = array(parameters.dens, dtype=complex64)
    f = array(parameters.f, dtype=complex64)
    re = 2.82e-13 * 1e2 / 1e-10
    n = 1 - dens * re * lamda**2 / 2 / pi * f * 1e-4
    n = toarray(n, code=complex64)
    sigmai = toarray(parameters.sigmai, code=float64)
    sigma = toarray(parameters.sigma, code=float64)
    sigmar = sqrt(sigma**2 - np.minimum(sigma, sigmai) ** 2)
    sigmar = sigmar[1:]
    sigmai = sigmai[1:] + 1e-5

    d = toarray(parameters.d, code=float64)
    d = r_[0, d[1:-1]]

    z = -cumsum(d)

    eta = sample.eta_x

    h = sample.h
    if h < 0 or h > 1:
        raise ValueError("The Hurst parameter h has to be between 0 and 1")

    eta_z = sample.eta_z

    (I, alpha, omega) = offspec.DWBA_Interdiff(
        qx, qz, lamda, n, z, sigmar, sigmai, eta, h, eta_z, d, taylor_n=instrument.taylor_n
    )
    I = real(I) * ((alphaR1 > 0) & (betaR1 > 0)) * instrument.I0

    # FootprintCorrections
    foocor = 1.0
    footype = instrument.footype
    beamw = instrument.beamw
    samlen = instrument.samplelen
    if footype == FootType.gauss:
        foocor = GaussIntensity(alphaR1, samlen / 2.0, samlen / 2.0, beamw)
    elif footype == FootType.square:
        foocor = SquareIntensity(alphaR1, samlen, beamw)
    elif footype == FootType.none:
        pass
    elif isinstance(footype, footprint_module.Footprint):
        foocor = footype(alphaR1, samlen)
    else:
        raise ValueError("The choice of footprint correction, footype," "is WRONG")
    # off-specular footprint correction of a homogeneous beam larger than the sample would be 1
    I *= foocor / alphaR1 * tth / 2.0

    restype = instrument.restype
    if restype == ResType.none:
        # if no resolution is defined, don't include specular peak
        return I * instrument.I0_offspec + instrument.Ibkg

    # include specular peak
    Ibgk = instrument.Ibkg
    instrument.Ibkg = 0.0
    # instrument.restype = ResType(0)
    if isinstance(tth, ndarray):
        Ispec = Specular(array(tth, dtype=float64), sample, instrument)
    else:
        Ispec = Specular(array([tth], dtype=float64), sample, instrument)[0]
    instrument.Ibkg = Ibgk
    # instrument.restype = restype

    if instrument.coords == Coords.tth:
        spec_peak = Ispec * exp(-0.5 * (TwoThetaQz / 2.0 - ThetaQx) ** 2 / instrument.detres**2)
    else:
        # angular resolution converted to qx grid
        qx_res = k * (cos(tth / 2) - cos(tth / 2 * (1 + instrument.detres / qz)))
        spec_peak = Ispec * exp(-0.5 * ThetaQx**2 / qx_res**2)

    # TODO: Make correct convolution of off-specular intensity and absolute scale
    return spec_peak + I * instrument.I0_offspec * instrument.detres * 1e3 + Ibgk


def SLD_calculations(z, item, sample: Sample, inst: Instrument):
    """Calculates the scatteringlength density as at the positions z

    # BEGIN Parameters
    z data.x
    item "Re"
    # END Parameters
    """
    parameters: LayerParameters = sample.resolveLayerParameters()
    dens = array(parameters.dens, dtype=complex64)
    f = array(parameters.f, dtype=complex64)
    sld = dens * f
    d_sld = sld[:-1] - sld[1:]
    d = array(parameters.d, dtype=float64)
    d = d[1:-1]
    # Include one extra element - the zero pos (substrate/film interface)
    int_pos = cumsum(r_[0, d])
    sigma = array(parameters.sigma, dtype=float64)[:-1] + 1e-7
    if z is None:
        z = arange(-sigma[0] * 5, int_pos.max() + sigma[-1] * 5, 0.5)
    rho = sum(d_sld * (0.5 - 0.5 * erf((z[:, newaxis] - int_pos) / sqrt(2.0) / sigma)), 1) + sld[-1]
    dic = {"Re": real(rho), "Im": imag(rho), "z": z, "SLD unit": "r_{e}/\\AA^{3}"}
    if item is None or item == "all":
        return dic
    else:
        try:
            return dic[item]
        except:
            raise ValueError("The chosen item, %s, does not exist" % item)


SimulationFunctions = {"Specular": Specular, "OffSpecular": OffSpecularMingInterdiff, "SLD": SLD_calculations}

Sample.setSimulationFunctions(SimulationFunctions)


class TestInterdiff(ModelTestCase):
    # TODO: currently this only checks for raise conditions in the code above, check of results should be added

    def test_spec(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, f=3e-5 + 1e-7j, dens=0.1)])],
            Ambient=Layer(),
            Substrate=Layer(f=5e-5 + 2e-7j, dens=0.1),
        )
        instrument = Instrument(
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            wavelength=1.54,
        )
        with self.subTest("x-ray tth"):
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray q"):
            instrument.coords = Coords.q
            Specular(self.qz, sample, instrument)

        # resolution corrections
        with self.subTest("x-ray q-res-fast"):
            instrument.restype = ResType.fast_conv
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-fast-var"):
            instrument.restype = ResType.fast_conv_var
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-full"):
            instrument.restype = ResType.full_conv_abs
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-gauss"):
            instrument.restype = resolution_module.GaussianResolution(sigma=0.001)
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-trapezoid"):
            instrument.restype = resolution_module.TrapezoidResolution(inner_width=0.001, outer_width=0.002)
            instrument.restype.get_weight_example()
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-wrong"):
            instrument.restype = 123
            with self.assertRaises(ValueError):
                Specular(self.qz, sample, instrument)
        instrument.restype = ResType.none

        # footprint corrections
        with self.subTest("x-ray q-footprint-square"):
            instrument.footype = FootType.square
            Specular(self.qz, sample, instrument)
        instrument.coords = Coords.tth
        with self.subTest("x-ray tth-footprint-square"):
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-gauss"):
            instrument.footype = FootType.gauss
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-gauss-offset"):
            instrument.footype = GaussianBeamOffset(sigma=0.5, offset=0.1)
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-square-offset"):
            instrument.footype = SquareBeamOffset(width=0.1, offset=0.05)
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-trapezoid"):
            instrument.footype = TrapezoidBeam(inner_width=0.1, outer_width=0.2)
            Specular(self.tth, sample, instrument)
            instrument.footype = TrapezoidBeam(inner_width=0.1, outer_width=0.1)
            Specular(self.tth, sample, instrument)
            instrument.footype = TrapezoidBeam(inner_width=0.0, outer_width=0.2)
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-wrong"):
            instrument.footype = 123
            with self.assertRaises(ValueError):
                Specular(self.qz, sample, instrument)

    def test_offspec(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, f=3e-5 + 1e-7j, dens=0.1)])],
            Ambient=Layer(),
            Substrate=Layer(f=5e-5 + 2e-7j, dens=0.1),
        )
        instrument = Instrument(
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            wavelength=1.54,
        )
        with self.subTest("x-ray tth"):
            OffSpecularMingInterdiff(self.tth, self.tth, sample, instrument)
        with self.subTest("x-ray q"):
            instrument.coords = Coords.q
            OffSpecularMingInterdiff(0.01 * self.qz, self.qz, sample, instrument)

        # resolution corrections
        # with self.subTest("x-ray q-res-fast"):
        #    instrument.restype = ResType.fast_conv
        #    OffSpecularMingInterdiff(0.01 * self.qz, self.qz, sample, instrument)
        #    OffSpecularMingInterdiff(0.001, self.qz, sample, instrument)
        with self.subTest("x-ray tth-res-fast"):
            instrument.coords = Coords.tth
            OffSpecularMingInterdiff(self.tth, self.tth, sample, instrument)
            OffSpecularMingInterdiff(0.1, self.tth, sample, instrument)
        instrument.restype = ResType.none
        instrument.coords = Coords.q

        # footprint corrections
        with self.subTest("x-ray q-footprint-square"):
            instrument.footype = FootType.square
            OffSpecularMingInterdiff(0.01 * self.qz, self.qz, sample, instrument)
        instrument.coords = Coords.tth
        with self.subTest("x-ray tth-footprint-square"):
            OffSpecularMingInterdiff(self.tth, self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-gauss"):
            instrument.footype = FootType.gauss
            OffSpecularMingInterdiff(self.tth, self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-gauss-offset"):
            instrument.footype = GaussianBeamOffset(sigma=0.5, offset=0.1)
            OffSpecularMingInterdiff(self.tth, self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-square-offset"):
            instrument.footype = SquareBeamOffset(width=0.1, offset=0.05)
            OffSpecularMingInterdiff(self.tth, self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-trapezoid"):
            instrument.footype = TrapezoidBeam(inner_width=0.1, outer_width=0.2)
            OffSpecularMingInterdiff(self.tth, self.tth, sample, instrument)
            instrument.footype = TrapezoidBeam(inner_width=0.1, outer_width=0.1)
            OffSpecularMingInterdiff(self.tth, self.tth, sample, instrument)
            instrument.footype = TrapezoidBeam(inner_width=0.0, outer_width=0.2)
            OffSpecularMingInterdiff(self.tth, self.tth, sample, instrument)

    def test_sld(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, f=3e-5 + 1e-7j, dens=0.1)])],
            Ambient=Layer(),
            Substrate=Layer(f=5e-5 + 2e-7j, dens=0.1),
        )
        instrument = Instrument(
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            wavelength=4.5,
        )
        with self.subTest("sld xray"):
            SLD_calculations(None, None, sample, instrument)


def standard_xray():
    """
    return the defied standard x-ray reflectivity to compare against other models
    """
    qz = linspace(0.01, 0.3, 15)
    return Specular(
        qz,
        Sample(
            Ambient=Layer(),
            Substrate=Layer(d=150.0, f=1e-5 + 1e-8j, dens=0.1),
            Stacks=[Stack(Layers=[Layer(d=150.0, f=2e-5 + 2e-8j, dens=0.1)])],
        ),
        Instrument(coords=Coords.q, wavelength=1.54, footype=FootType.none, restype=ResType.none),
    )



# Update the genx version in .iss file
import sys
import os

folder=os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(folder))
from genx import version

txt=open(os.path.join(folder, 'genx_template.iss'), 'r').read()
txt=txt.replace('{version}', version.__version__)
open(os.path.join(folder, 'genx.iss'), 'w').write(txt)



"""
===================================
:mod:`seifert_nja` NJA ASCII format
===================================

Loads the data from *.nja ASCII files with limited header analysis.
"""

import numpy as np

from ..data_loader_framework import Template
from ..utils import ShowWarningDialog

try:
    import wx
except ImportError:
    # in case of console usinge withou wx beeing installed put a mock class/module
    class void:
        pass

    wx = void()
    wx.Dialog = void


class Plugin(Template):
    wildcard = "*.nja"

    def __init__(self, parent):
        Template.__init__(self, parent)

    def CanOpen(self, file_path):
        if not Template.CanOpen(self, file_path):
            return False
        with open(file_path, "r", encoding="utf-8") as fh:
            l1 = fh.readline()
            l2 = fh.readline()
        return l2.startswith("#ScanTableParameter") or l1.startswith('#MeasParameter')

    def evaluate_header(self, header):
        output = {}
        for line in header:
            if not line.startswith("&"):
                continue
            items = line.split("&")
            if items[1].startswith("ScanAxis"):
                output["ScanAxis"] = items[1].split("=")[1].strip()
            elif items[1].startswith("Axis"):
                output[items[1]] = dict([tuple(ei.strip().split("=", 2)) for ei in items[1:]])
            else:
                for ei in items[1:]:
                    output[ei.split("=")[0]] = ei.split("=", 2)[-1].strip()

        return output

    def LoadData(self, dataset, filename, data_id=0):
        """LoadData(self, dataset, filename) --> none

        Loads the data from filename into the dataset object.
        """
        try:
            with open(filename, encoding="utf-8", errors="ignore") as fh:
                header = [fh.readline()]
                while not header[-1].startswith("&NoValues="):
                    header.append(fh.readline())
                load_array = np.loadtxt(fh, delimiter=None)
        except Exception as e:
            ShowWarningDialog(
                self.parent,
                "Could not load the file: "
                + filename
                + " \nPlease check the format.\n\n numpy.loadtxt"
                + " gave the following error:\n"
                + str(e),
            )
        else:
            # For the freak case of only one data point
            if len(load_array.shape) < 2:
                load_array = np.array([load_array])
            # Check so we have enough columns
            if load_array.shape[1] < 2:
                ShowWarningDialog(
                    self.parent,
                    "The data file does not contain enough number of columns. It has "
                    + str(load_array.shape[1])
                    + " columns. Rember that the column index start at zero!",
                )
                self.SetStatusText("Could not load data - not enough columns")
                return

            # The data is set by the default Template.__init__ function
            # Know the loaded data goes into *_raw so that they are not
            # changed by the transforms
            dataset.x_raw = load_array[:, 0]
            dataset.y_raw = load_array[:, 1]

            header = self.evaluate_header(header)
            if header.get("ScanAxis", "") == "O":
                dataset.x_command = "2*x"
            else:
                dataset.x_command = "x"

            # Check if we have errors in the data - if not handle it with nan's
            dataset.error_raw = np.sqrt(load_array[:, 1])

            # Run the commands on the data - this also sets the x,y, error members of the data item.
            dataset.run_command()

            # insert metadata into ORSO compatible fields
            dataset.meta["data_source"]["facility"] = "Seifert nja loader"
            dataset.meta["data_source"]["experiment"]["probe"] = "x-ray"
            dataset.meta["data_source"]["measurement"]["scheme"] = "angle-dispersive"
            # header evaluations
            inst = dataset.meta["data_source"]["measurement"]["instrument_settings"]
            inst["incident_angle"] = {
                "min": float(dataset.x_raw.min()),
                "max": float(dataset.x_raw.max()),
                "unit": "degree",
            }
            inst["wavelength"] = {"magnitude": 1.54, "unit": "angstrom"}
            # insert all metadata into user field
            dataset.meta["nja_header"] = header



"""
===========================================
:mod:`bruker`  Bruker format (.brml) reader
===========================================

Loads the data exported by e.g. Bruker D8 instruments.
"""

import io
import xml.etree.ElementTree as ET
import zipfile

import numpy as np

from ..data_loader_framework import Template
from ..utils import ShowWarningDialog


class Plugin(Template):
    wildcard = "*.brml"

    def __init__(self, parent):
        Template.__init__(self, parent)

    def get_datasets(self, zip_ref):
        # parse the DataContainer.xml file to get a list of all measurements.
        data_container = ET.fromstring(zip_ref.open("Experiment0/DataContainer.xml", "r").read())
        return [ds.text for ds in data_container.find("RawDataReferenceList").iter("string")]

    def CountDatasets(self, file_path):
        try:
            with zipfile.ZipFile(file_path, "r") as zip_ref:
                return len(self.get_datasets(zip_ref))
        except Exception as e:
            return 1

    def get_columns(self, raw_data_views: ET.Element):
        columns = []
        for raw_data_view in raw_data_views:

            if "LogicName" in raw_data_view.attrib and raw_data_view.attrib["Length"] == "1":
                columns.append(raw_data_view.attrib["LogicName"])
            elif raw_data_view.attrib["Length"] == "1":
                if raw_data_view[0].attrib.get("Category", None) == "Count":
                    columns.append("Count")
                else:
                    columns.append(raw_data_view[0].attrib["LogicName"])
            else:
                for fd in raw_data_view.iter("FieldDefinitions"):
                    if "AxisId" in fd.attrib:
                        columns.append(fd.attrib["AxisId"])
                    else:
                        columns.append(fd.attrib["FieldName"])
        return columns

    def LoadData(self, dataset, filename, data_id=0):
        """
        Loads the data from filename into the data_item_number.
        """
        try:
            with zipfile.ZipFile(filename, "r") as zip_ref:
                # get a list of datasets in this file
                datasets = self.get_datasets(zip_ref)
                data_to_load = datasets[data_id]
                data_xml = ET.fromstring(zip_ref.open(data_to_load, "r").read())
            data_lines = [dt.text for dt in data_xml.find("DataRoutes").find("DataRoute").iter("Datum")]
            load_array = np.loadtxt(io.StringIO("\n".join(data_lines)), delimiter=",")
            columns = self.get_columns(data_xml.find("DataRoutes").find("DataRoute").find("DataViews"))
        except Exception as e:
            ShowWarningDialog(
                self.parent,
                "Could not load the file: "
                + filename
                + " \nPlease check the format.\n\n numpy.loadtxt / header analysis "
                + " gave the following error:\n"
                + str(e),
            )
        else:
            try:
                x_idx = columns.index("Theta")
                tth_only = False
            except ValueError:
                x_idx = columns.index("TwoTheta")
                tth_only = True
            y_idx = columns.index("Count")
            try:
                att_idx = columns.index("AbsorptionFactor")
                atten = load_array[:, att_idx]
            except ValueError:
                att_idx = None
                atten = 1.0

            dataset.x_raw = load_array[:, x_idx]
            dataset.y_raw = load_array[:, y_idx]
            dataset.error_raw = np.sqrt(load_array[:, y_idx]*atten)
            if not tth_only:
                dataset.x_command = "2*x"
            else:
                dataset.x_command = "x"
            if att_idx is not None:
                # measurement reports attenuators
                dataset.set_extra_data("atten", atten)
            # Run the commands on the data - this also sets the x,y, error memebers
            # of that data item.
            dataset.run_command()

            # insert metadata into ORSO compatible fields
            try:
                info_data = list(data_xml.find("FixedInformation").find("MethodInformation").iter("InfoData"))
            except AttributeError:
                info_data = []
                info_list = []
            for id in info_data:
                try:
                    wavelength = float(id.find("MethodAlignment").find("WaveLength").attrib["Value"])
                except (AttributeError, ValueError):
                    wavelength = 1.54
                try:
                    info_list = list(id.find("InfoList").iter())
                except (AttributeError, ValueError):
                    pass
            for il in info_list:
                if il.attrib.get("Name", "") == "SampleName":
                    dataset.meta["data_source"]["sample"] = {"name": il.attrib["Value"]}
                if il.attrib.get("Name", "") == "User":
                    dataset.meta["data_source"]["owner"]["name"] = il.attrib["Value"]
            dataset.meta["data_source"]["experiment"]["instrument"] = "Bruker"
            dataset.meta["data_source"]["experiment"]["probe"] = "x-ray"
            try:
                dataset.meta["data_source"]["experiment"]["start_date"] = data_xml.find("TimeStampStarted").text
            except AttributeError:
                pass
            dataset.meta["data_source"]["measurement"]["scheme"] = "angle-dispersive"
            inst = dataset.meta["data_source"]["measurement"]["instrument_settings"]
            inst["incident_angle"] = {
                "min": float(dataset.x_raw.min()),
                "max": float(dataset.x_raw.max()),
                "unit": "deg",
            }
            inst["wavelength"] = {
                "magnitude": wavelength,
                "unit": "angstrom",
            }
